/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "wfs_cache_blockfile.h"
#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>

#include <wfs_misc.h>

wfs_cache_blockfile::wfs_cache_blockfile( std::string cache_dir ) :
  _cache_dir(cache_dir), MAX_PERFILE_BUF_SIZE(1024*1024*100) {

  // make sure it exists
  struct stat st;
  if( stat( _cache_dir.c_str(), &st ) ) {
    wfs_log(WFS_DBG_OFF, "The cache directory %s does not exist.  Please "
             "create it or specify another.\n", _cache_dir.c_str() );
    exit(-1);
  }

  assert( pthread_mutex_init( &_writes_lock, NULL ) == 0 );
  assert( pthread_mutex_init( &_dirs_lock, NULL ) == 0 );
  assert( pthread_mutex_init( &_readdirs_lock, NULL ) == 0 );
  assert( pthread_mutex_init( &_md_lock, NULL ) == 0 );

}

wfs_cache_blockfile::~wfs_cache_blockfile() {
  // TODO: delete existing buffered writes?
  pthread_mutex_destroy( &_writes_lock );
  pthread_mutex_destroy( &_dirs_lock );
  pthread_mutex_destroy( &_readdirs_lock );
  pthread_mutex_destroy( &_md_lock );
}

std::string wfs_cache_blockfile::filename( wfs_file_handle fh,
                                           wfs_version version,
                                           unsigned int block_no ) {

  char buf[_cache_dir.length()+100];
  sprintf( buf, "%s/%016"PRIx64".%d.%016"PRIx64".%d", _cache_dir.c_str(),
           file_handle_to_file_id(fh), version.version_num, version.node_id,
	   block_no );
  return std::string( buf, strlen(buf) );
}

std::string wfs_cache_blockfile::filename_md( wfs_file_handle fh ) {

  char buf[_cache_dir.length()+100];
  sprintf( buf, "%s/%016"PRIx64".md", _cache_dir.c_str(),
           file_handle_to_file_id(fh) );
  return std::string( buf, strlen(buf) );
}

bool wfs_cache_blockfile::read_md( wfs_file_handle fh,
                                   wfs_protocol::fileinfo &finfo ) {

  wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);
  pthread_mutex_lock( &_md_lock );
  bool ans = false;
  wfs_protocol::fileinfo *md = _md[ fid ];
  if( md != NULL ) {
    finfo = *md;
    ans = true;
  }
  pthread_mutex_unlock( &_md_lock );
  return ans;
}

bool wfs_cache_blockfile::read_any_md( wfs_file_handle fh,
                                       wfs_protocol::fileinfo &finfo ) {

  if( !read_md( fh, finfo ) ) {
    // do we have a saved version?
    return readin_md( fh, finfo );
  } else {
    return true;
  }

}

void wfs_cache_blockfile::store_md( wfs_file_handle fh,
                                    wfs_protocol::fileinfo finfo ) {

  wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);
  pthread_mutex_lock( &_md_lock );
  wfs_protocol::fileinfo *md = _md[ fid ];
  if( md == NULL ) {
    md = new wfs_protocol::fileinfo();
    _md[fid] = md;
  }

  *md = finfo;
  assert( _md[fid]->size == finfo.size );

  wfs_log(WFS_DBG_2, "Storing md for file %016"PRIx64" with size %lld (%lld)\n",
	    fh.fid, finfo.size, _md[fid]->size );
  pthread_mutex_unlock( &_md_lock );

}

int md_ver_size = 2*(sizeof(wfs_version) + sizeof(unsigned long long));

void wfs_cache_blockfile::writeout_md( wfs_file_handle fh,
                                       wfs_version version ) {

  wfs_protocol::fileinfo finfo;
  if( !read_md( fh, finfo ) ) {
    return;
  }

  std::string filename = filename_md( fh );

  int fd = creat( filename.c_str(), 0666 );
  if( fd >= 0 ) {

    // order:
    // 4-byte version
    // 8-byte size
    // permanent cues

    char ver[md_ver_size+1];
    sprintf( ver, "%08x%016"PRIx64"%016llx", version.version_num, version.node_id, finfo.size );
    int n = write( fd, ver, strlen(ver) );
    std::string perm_cues = fh.perm_cues_string();
    n = write( fd, perm_cues.c_str(), perm_cues.length() );

    close( fd );

  }

}

bool wfs_cache_blockfile::readin_md( wfs_file_handle fh,
                                     wfs_protocol::fileinfo &finfo ) {

  std::string filename = filename_md( fh );

  int fd = open( filename.c_str(), O_RDONLY );
  if( fd >= 0 ) {

    char ver[md_ver_size+1];
    int n = read( fd, ver, md_ver_size );
    if( n != md_ver_size ) {
      close( fd );
      return false;
    }
    ver[md_ver_size] = '\0';
    int32_t v;
    uint32_t id1;
    uint32_t id2;
    uint32_t s1;
    uint32_t s2;
    n = sscanf( ver, "%08x%08x%08x%08x%08x", &v, &id1, &id2, &s1, &s2 );
    if( n != 3 ) {
      close( fd );
      return false;
    }
    finfo.version.version_num = v;
    finfo.version.node_id = two_ints_to_node_id( id1, id2 );
    finfo.size = two_ints_to_node_id( s1, s2 ); // TODO: not really a node ID

    char cues[1000];
    n = read( fd, cues, 1000 );
    finfo.perm_cues = std::string( cues, n );

    close( fd );
    wfs_log(WFS_DBG_1, "Reading in meta-data for %016"PRIx64", version=%d, node=%"PRIu64", size=%llu\n",
              fh.fid, finfo.version.version_num, finfo.version.node_id,finfo.size );
    return true;

  } else {
    return false;
  }

}

void wfs_cache_blockfile::clear_md( wfs_file_handle fh, wfs_version version ) {


  wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);
  pthread_mutex_lock( &_md_lock );
  wfs_protocol::fileinfo *md = _md[ fid ];
  if( md != NULL ) {
    delete md;
  }
  _md[fid] = NULL;
  pthread_mutex_unlock( &_md_lock );

}

// how many writes have we had to this version of the file?
int wfs_cache_blockfile::num_buf_writes( wfs_file_handle fh ) {

  pthread_mutex_lock( &_writes_lock );

  wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);
  buf_write_entry *entry = _writes[fid];
  int i = 0;
  while( entry != NULL ) {
    i++;
    entry = entry->next;
  }

  pthread_mutex_unlock( &_writes_lock );

  return i;

}

void wfs_cache_blockfile::apply_buf_writes( wfs_file_handle fh,
                                            wfs_version version,
                                            unsigned int block_no,
                                            std::string &buf ) {

  unsigned int n = buf.length();

  pthread_mutex_lock( &_writes_lock );

  wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);
  buf_write_entry *entry = _writes[fid];

  // apply every write affecting this block
  while( entry != NULL ) {

    wfs_log(WFS_DBG_3, "offset: %llu\n",entry->write.offset);
    wfs_log(WFS_DBG_3, "entry version: %d, %016"PRIx64"\n", entry->version.version_num, entry->version.node_id);
    wfs_log(WFS_DBG_3, "version: %d, %016"PRIx64"\n", version.version_num, version.node_id);

    if( entry->version >= version ) {
      unsigned int num_blocks;
      unsigned int start_block_no = calc_block_no( entry->write.offset,
                                                   entry->write.buf.length(),
                                                   num_blocks );

      if( start_block_no <= block_no &&
          (block_no < (start_block_no + num_blocks) ||
           entry->write.truncate) ) {

        // begin complicated algorithm to figure out where in the block
        // the write should occur, and how big it should be.

        int owb = offset_within_block( block_no, entry->write.offset );
        int oww = 0;
        unsigned copy_length = entry->write.buf.length();

        if( owb < 0 ) {
          copy_length += owb;
          oww -= owb;
          owb = 0;
        }

        // it's sufficient to truncate the length relative to the
        // offset-within-block, rather than also truncating relative
        // to the offset-within-write, since the copy_length can never
        // be bigger than the buffered write's buffer, and so there's
        // no need to truncate based on the offset within that buffer;
        // all that matters is whether the amount copied over will fit
        // inside the block being read.

	if( (owb + copy_length) > WFS_BLOCK_SIZE ) {
          copy_length = WFS_BLOCK_SIZE - owb;
        }

        if( entry->write.truncate ) {

          // a truncate just needs to set the effective length of the buffer
          // to the offset within the block
          // because of the enclosing if statement, this will apply to
          // any block including or after the point of truncation
          wfs_log(WFS_DBG_1, "Applying truncate to size %d in block %d\n", owb,
                    block_no );
          n = owb;

        } else if( copy_length > 0 ) {

          wfs_log(WFS_DBG_1, "Applying buffered write at %d in block %d, length %d, "
                    "from offset within write of %d\n",
                    owb, block_no, copy_length, oww );

          assert(owb >= 0);
          assert(copy_length >= 0 && owb + copy_length <= WFS_BLOCK_SIZE);
          assert(oww >= 0);
          assert(oww + copy_length <= entry->write.buf.length());

          if( owb + copy_length > buf.length() ) {
            buf.resize( owb + copy_length, '\0' );
          }

          buf.replace( owb, copy_length, entry->write.buf, oww, copy_length );

          if( owb + copy_length > n ) {
            n = owb + copy_length;
          }

        }
      } else if( start_block_no > block_no ) {
        // this means that there was a write past this block in the file,
        // which means that this whole block should exist, even if it's
        // just all zeros.
        n = WFS_BLOCK_SIZE;

      }
    }

    entry = entry->next;
  }

  pthread_mutex_unlock( &_writes_lock );

  if( n != buf.length() ) {
    buf.resize( n, '\0' );
  }

}

bool wfs_cache_blockfile::read_block( wfs_file_handle fh, wfs_version version,
                                      unsigned int block_no ) {

  std::string f = filename( fh, version, block_no );
  struct stat sb;
  bool x = (stat( f.c_str(), &sb ) == 0);
  if (x) {
    wfs_log(WFS_DBG_1, "Read block mem true for %016"PRIx64":%d:%d\n",
	      fh.fid, version.version_num, block_no );
  }
  return x;
}

bool wfs_cache_blockfile::read_block( wfs_file_handle fh, wfs_version version,
                                      unsigned int block_no,
                                      std::string &buf ) {

  std::string f = filename( fh, version, block_no );
  int fd = open( f.c_str(), O_RDONLY );

  if( fd <= 0 ) {
    return false;
  }

  /**
   // Uncomment this code to read directly into the string's buffer,
   // in order to avoid a copy.  Seems to be (very slightly) faster.
   // However, there is no guarantee that buf[0] points to a
   // continuous block of WFS_BLOCK_SIZE bytes -- string is allowed
   // to break its buffer into separate pieces, and combine if
   // data() is called.  Thus, this code is technically incorrect,
   // but I think it would work in almost all cases.
   //
   // Other ways of writing direct to a string (ifstream iterators,
   // stringstreams, etc) seem to be much slower.
   buf.resize( WFS_BLOCK_SIZE );
   int n = read( fd, &(buf[0]), WFS_BLOCK_SIZE );
   buf.resize(n);
  */

  // comment out this block of code if using the above string-direct way
  char read_buf[WFS_BLOCK_SIZE];
  int n = 1;
  bzero( read_buf, WFS_BLOCK_SIZE );
  n = read( fd, read_buf, WFS_BLOCK_SIZE );
  buf = std::string( read_buf, n );

  close( fd );

  apply_buf_writes( fh, version, block_no, buf );

  return true;

}


void wfs_cache_blockfile::store_block( wfs_file_handle fh, wfs_version version,
                                       unsigned int block_no,
                                       std::string buf ) {

  std::string f = filename( fh, version, block_no );
  int fd = creat( f.c_str(), 0666 );

  // TODO: might need some locks here to make sure two threads don't
  // try to create the same file at the same time, though methinks this
  // unlikely

  if( fd <= 0 ) {
    wfs_log(WFS_DBG_OFF, "Couldn't create cache file %s\n", f.c_str() );
  }

  int n = write( fd, buf.c_str(), buf.length() );
  assert (n != -1);

  close( fd );

  writeout_md( fh, version );

}

bool wfs_cache_blockfile::buf_write( wfs_file_handle fh, wfs_version version,
                                     unsigned long long offset,
                                     bool truncate,
                                     const std::string &buf ) {

  // TODO: would be better to have a separate lock for each file.  Oh well.
  pthread_mutex_lock( &_writes_lock );

  wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);

  buf_write_entry *first = _writes[fid];
  buf_write_entry *last = _last_write[fid];

  // TODO: this is sort of a hack, since we should really be keeping track
  // the total size among all files.  But then the wfs_client won't know
  // which files to buffer without widening the API . . .
  if( (last != NULL) && (last->total_size > MAX_PERFILE_BUF_SIZE) ) {
    pthread_mutex_unlock( &_writes_lock );
    return false;
  }

  buf_write_entry *entry = new buf_write_entry();
  entry->next = NULL;
  entry->version = version;
  entry->write.offset = offset;
  entry->write.buf = buf;
  entry->total_size = buf.size();
  entry->write.truncate = truncate;

  if( last != NULL ) {
    // append to list, so that the writes are stored in order
    last->next = entry;
    entry->total_size += last->total_size;
  }

  _last_write[fid] = entry;

  if( first == NULL ) {
    _writes[fid] = entry;
  }

  pthread_mutex_unlock( &_writes_lock );
  return true;

}

void wfs_cache_blockfile::get_buf_writes( wfs_file_handle fh,
                                          std::vector<wfs_buf_write> &buf_writes ) {

  // TODO: we're just returning all buffered writes for any version.
  // shouldn't be a big deal, but maybe there's a better way to do this.

  pthread_mutex_lock( &_writes_lock );

  wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);

  buf_write_entry *entry = _writes[fid];

  while( entry != NULL ) {
    buf_writes.push_back( entry->write );

    buf_write_entry *x = entry;
    entry = entry->next;
    delete x;
  }

  _writes[fid] = NULL;
  _last_write[fid] = NULL;

  pthread_mutex_unlock( &_writes_lock );

}

bool wfs_cache_blockfile::dir_lookup( wfs_file_handle dh,
                                      std::string name,
                                      wfs_file_handle &fh ) {

  bool found = false;

  pthread_mutex_lock( &_dirs_lock );

  if( _dirs.find(dh.fid) != _dirs.end() &&
      _dirs[dh.fid].find(name) != _dirs[dh.fid].end() ) {
    fh.fid = _dirs[dh.fid][name];
    found = true;
  }

  // negative caching
  if( !found && _complete_dirs[dh.fid] == true ) {
    fh.fid = 0;
    found = true;
  }

  pthread_mutex_unlock( &_dirs_lock );

  return found;

}

void wfs_cache_blockfile::add_dir_entry( wfs_file_handle dh, std::string name,
                                         wfs_file_handle fh ) {

  pthread_mutex_lock( &_dirs_lock );

  _dirs[dh.fid][name] = fh.fid;
  // initialize just in case
  if( _complete_dirs.find(dh.fid) == _complete_dirs.end() ) {
    _complete_dirs[dh.fid] = false;
  }

  bool complete = _complete_dirs[dh.fid];

  pthread_mutex_unlock( &_dirs_lock );

  // we only really care about updating the readdir entry if we have a complete
  // directory; otherwise readdir should probably just go to the server
  // regardless
  // TODO: make this faster
  if( complete ) {
    pthread_mutex_lock( &_readdirs_lock );

    // placeholder if we need it

    // add it to the readdir entries if it doesn't exist yet
    // or delete if needed
    bool in_readdir = false;
    std::vector<wfs_storage::dirent>::iterator i;
    for( i = _readdirs[dh.fid].begin(); i != _readdirs[dh.fid].end(); i++ ) {
      if( (*i).name == name ) {
	in_readdir = true;
	break;
      }
    }

    // TODO: This could possibly mess up listing directories that
    // take multiple RPCs, since this may put the new guy in the wrong
    // place in the order.  Maybe make it alphabetical??
    if( fh.fid != 0 ) {

      if( !in_readdir ) {
	// make some room
	_readdirs[dh.fid].resize( _readdirs[dh.fid].size()+1 );
	i = _readdirs[dh.fid].end()-1;
      }

      (*i).inum = fh;
      (*i).name = name;

    } else if( in_readdir && fh.fid == 0 ) {
      // delete from readdir if this is negative caching
      _readdirs[dh.fid].erase(i);
    }
  }

  pthread_mutex_unlock( &_readdirs_lock );

}

bool wfs_cache_blockfile::readdir( wfs_file_handle dh, unsigned long long ind,
                                   unsigned int n,
                                   std::vector<wfs_storage::dirent> &e ) {

  bool found = false;

  pthread_mutex_lock( &_readdirs_lock );

  if( _readdirs.find(dh.fid) != _readdirs.end() ) {
    std::vector<wfs_storage::dirent> v = _readdirs[dh.fid];
    // if the highest index read is bigger than the requested index,
    // than we have at least done a fetch covering that index.
    // Assume that this means we have fetched everything for beyond that index
    if( v.size() > 0 && _highest_readdir[dh.fid] > ind ) {
      found = true;
      if( ind + n <= v.size() ) {
        e.resize( n );
      } else {
        e.resize( v.size() - ind );
      }
      for( unsigned i = 0; i < n && i+ind < v.size(); i++ ) {
        e[i] = v[i+ind];
      }
    } else {
      pthread_mutex_lock( &_dirs_lock );
      if( (_complete_dirs.find(dh.fid) != _complete_dirs.end()) &&
	  (_complete_dirs[dh.fid] == true) ) {
	found = true;
	for (unsigned i = ind; i < v.size(); i++) {
	  e.push_back(v[i]);
	  if (e.size() == n)
	    break;
	}
      }
      pthread_mutex_unlock( &_dirs_lock );
    }

  }

  pthread_mutex_unlock( &_readdirs_lock );

  return found;

}

void wfs_cache_blockfile::add_readdir_entry( wfs_file_handle dh,
                                             unsigned long long ind,
                                             std::vector<wfs_storage::dirent> e,
                                             bool complete ) {

  pthread_mutex_lock( &_readdirs_lock );
  pthread_mutex_lock( &_dirs_lock );

  if( _readdirs[dh.fid].size() < ind + e.size() ) {
    _readdirs[dh.fid].resize( ind + e.size() );
  }

  // keep track of the largest index we've fetched
  if( _highest_readdir[dh.fid] < ind+1 ) {
    _highest_readdir[dh.fid] = ind+1;
  }

  unsigned i;
  for( i = 0; i < e.size(); i++ ) {
    _readdirs[dh.fid][i+ind] = e[i];
    _dirs[dh.fid][e[i].name] = e[i].inum.fid;
  }

  _complete_dirs[dh.fid] = complete;

  pthread_mutex_unlock( &_dirs_lock );
  pthread_mutex_unlock( &_readdirs_lock );

}

void wfs_cache_blockfile::flush_dir( wfs_file_handle dh ) {

  pthread_mutex_lock( &_readdirs_lock );
  pthread_mutex_lock( &_dirs_lock );

  _dirs[dh.fid].clear();
  _readdirs[dh.fid].clear();
  _readdirs[dh.fid].resize(0);
  _complete_dirs[dh.fid] = false;
  _highest_readdir[dh.fid] = 0;

  pthread_mutex_unlock( &_dirs_lock );
  pthread_mutex_unlock( &_readdirs_lock );

}
