/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */



#include <wfs_misc.h>
#include <method_thread.h>
#include "wfs_bt_client.h"
#include <fcntl.h>
#include <algorithm>

const unsigned wfs_bt_client::ASK_LIST_SIZE = 5;

wfs_bt_client::wfs_bt_client( sockaddr_in bt_addr, wfs_node me,
			      slice_manager *slices,
			      wfs_cache *cache, rpcs *rpc_server,
			      vivaldi *viv, wfs_client_manager *clients,
			      bool read_ahead, int prefetch_threads ) :
  wfs_cache_client( bt_addr, me, slices, cache, rpc_server, viv, clients,
		    read_ahead, prefetch_threads ),
  _bt_addr(bt_addr){

  pthread_mutex_init( &_nodes_in_use_lock, NULL );
  pthread_mutex_init( &_my_block_map_lock, NULL );
  pthread_mutex_init( &_acs_lock, NULL );
  pthread_cond_init( &_acs_condvar, NULL );

  rpc_server->ident_reg( wfs_protocol::readfile, this,
			 &wfs_bt_client::readcachedfile );

  // fork to prevent blocking the data on a slow trip to the
  // origin server
  method_thread( this, true, &wfs_bt_client::add_cache_loc );

}

int wfs_bt_client::readcachedfile( const std::string &user,
				   wfs_file_handle fh, wfs_version version,
				   unsigned long long off, unsigned int n,
				   wfs_protocol::readinfo &info ) {

  wfs_log(WFS_DBG_1, "wfs_bt_client::getcachedfile %016"PRIx64" off=%lld n=%d\n", fh.fid, off, n );
  assert((off % WFS_BLOCK_SIZE) == 0);
  assert(n == WFS_BLOCK_SIZE);
  int block_no = off/WFS_BLOCK_SIZE;

  // permissions checking
  wfs_protocol::fileinfo finfo;
  bool success = _cache->read_any_md(fh,finfo);
  if (!success) {
    wfs_log(WFS_DBG_1, "wfs_bt_client::getcachedfile couldn't check permissions!");
    return wfs_protocol::PERM;
  } else if (!has_permission(finfo.permissions, user, WFS_PERMISSION_READ)) {
    return wfs_protocol::PERM;
  }

  if( !_cache->read_block( fh, version, block_no, info.buf ) ) {
    return wfs_protocol::IOERR;
  } else {
    pthread_mutex_lock( &_my_block_map_lock );
    info.block_map = _my_block_map[fh.fid];
    pthread_mutex_unlock( &_my_block_map_lock );

    // also pick out some other nodes we know about that have this block,
    // and pass along that info
    _cache_loc.get_rand_nodes( fh.fid, version, off, n, info.others,
			       ASK_LIST_SIZE );

    wfs_log(WFS_DBG_1, "BT-sending %d block map, %d list size\n", info.block_map.size(), info.others.size() );

    return wfs_protocol::OK;
  }
}

wfs_protocol::status
wfs_bt_client::read_uncached_data( wfs_file_handle fh,
				   wfs_version version,
				   unsigned long long off,
				   unsigned int n,
				   const std::string &user,
				   std::string &buf ) {

  if( fh.get_cue( WFS_CUE_POPULARITY ).val != WFS_CUE_HOTSPOT ) {
    return wfs_cache_client::read_uncached_data( fh, version, off, n, user, buf );
  }

  wfs_file_id fid = file_handle_to_file_id( fh );
  wfs_node file_node = get_primary( fid );
  if( file_node.id == 0 ) {
    return wfs_protocol::IOERR;
  }

  wfs_log(WFS_DBG_1, "bt:read_uncached_data: %016"PRIx64",%d (%llu, %d) file_node %s:%d\n",
	    fh.fid, version.version_num, off, n,
	    inet_ntoa(file_node.addr.sin_addr), ntohs(file_node.addr.sin_port));

  wfs_protocol::blockinfo binfo;
  binfo.off = off;
  binfo.size = n;
  binfo.version = version;

  sockaddr_in addr;
  addr.sin_family = AF_INET;

  bool nomiss = false;
  Coord cc; //space filler
  std::vector<cacheloc_ent> list;
  wfs_protocol::status status;

  time_t now = time(NULL);
  time_t then = now;
  if( _fetched_at_times.find(fid) != _fetched_at_times.end() ) {
    then = _fetched_at_times[fid];
  }

  // make sure there's at least one node, and it's not just the file
  // server
  bool cached_already = false;
  _cache_loc.get_close_nodes( fid, version, _viv, cc,
			      off, n, list, ASK_LIST_SIZE );
  for( unsigned i = 0; i < list.size(); i++ ) {
    if( list[i].s_addr != file_node.addr.sin_addr.s_addr ) {
      cached_already = true;
      break;
    }
  }

  while( _me.id != file_node.id &&
	 ( ((now-then > 60) && (list.size() < ASK_LIST_SIZE)) ||
	   !cached_already) ) {

    assert(!nomiss);
    list.clear();

    wfs_log(WFS_DBG_1, "Asking file node %016"PRIx64" for locations for file %016"PRIx64", data (%llu,%d) -- now-then=%ld already_cached=%d\n",
	      file_node.id, fh.fid, off, n, now-then, cached_already );

    wfs_log(WFS_DBG_1,"get_client temp "__FILE__".%d: %s\n",
	    __LINE__, inet_ntoa(file_node.addr.sin_addr));
    rpcc *client = _clients->get_client( file_node, fhto(fh) );
    if( client == NULL ) {
      status = rpc_const::timeout_failure;
    } else {
      status = client->call( wfs_protocol::getcacheloc,
			     fh, binfo, _viv->my_coords(), ASK_LIST_SIZE,
			     list, fhto(fh), "host" );
      _clients->done_with_client(file_node);
    }

    wfs_log(WFS_DBG_1, "Finished asking file node %016"PRIx64" for locations for file %016"PRIx64", data (%llu,%d) -- now-then=%ld already_cached=%d\n",
	      file_node.id, fh.fid, off, n, now-then, cached_already );

    if (status != wfs_protocol::OK) { // XXX: should not happen normally
      addr = file_node.addr;
      break;
    }

    now = then = time(NULL);
    _fetched_at_times[fid] = now;

    if (list.size() < ASK_LIST_SIZE) {
      wfs_log(WFS_DBG_2, "Got %u<%d cachelocs\n",list.size(),ASK_LIST_SIZE);
      cacheloc_ent e;
      e.s_addr = file_node.addr.sin_addr.s_addr;
      e.s_port = file_node.addr.sin_port;
      e.off.push_back(0);
      e.n.push_back( (unsigned long long) -1 );
      _viv->get_coords( file_node.addr.sin_addr.s_addr, e.c );
      list.push_back(e);
    }

    for (unsigned i = 0; i < list.size(); i++) {
      _viv->update_other_coords( list[i].s_addr, list[i].s_port,
				list[i].c, list[i].age );
    }

    if( wfs_dbg_block(WFS_DBG_1) ) {
      wfs_log(WFS_DBG_1, "choosing node out of list: " );
      for( unsigned i = 0; i < list.size(); i++ ) {
        sockaddr_in addr_x;
        addr_x.sin_addr.s_addr = list[i].s_addr;
        wfs_log(WFS_DBG_1_NOHEADER, "\t\t%s:%d (%s -- %.3f) ", 
		inet_ntoa(addr_x.sin_addr),
            ntohs(list[i].s_port),
            list[i].c.print_coord().c_str(),
            _viv->my_coords().distance_f(list[i].c) );
      }
      wfs_log(WFS_DBG_1_NOHEADER, "whereas my list of coords are %s\n",
          _viv->print_allcoords().c_str() );
    }
    _cache_loc.update_cacheloc( fid, version, list );
    list.clear();
    nomiss = true;
    cached_already = true;
  }

  // get a few more choices just in case
  list.clear();
  _cache_loc.get_close_nodes( fid, version, _viv, cc,
			      off, n, list, ASK_LIST_SIZE );


  // figure out the distance ratios between the nodes, which tells
  // us how many outstanding RPCs we prefer to have to the nodes
  int limit[list.size()];
  for( unsigned i = 0; i < list.size(); i++ ) {
    limit[i] = (int) ( _viv->my_coords().distance_f(list[list.size()-1].c)/
		       _viv->my_coords().distance_f(list[i].c));
    if( wfs_dbg_block(WFS_DBG_1) ) {
      sockaddr_in addr_x;
      addr_x.sin_addr.s_addr = list[i].s_addr;
      wfs_log(WFS_DBG_1, "%s:%d has limit %d (%.2f vs. %.2f)\n",
	      inet_ntoa(addr_x.sin_addr),
	      ntohs(list[i].s_port), limit[i],
	      _viv->my_coords().distance_f(list[i].c),
	      _viv->my_coords().distance_f(list[list.size()-1].c) );
    }
  }

  wfs_protocol::readinfo info;
  unsigned i = 0;
  do {
    cacheloc_ent *curr = NULL;
    bool fromclient;
    if( list.size() > i && _me.id != file_node.id ) {
      curr = &list[i];
      addr.sin_addr.s_addr = curr->s_addr;
      addr.sin_port = curr->s_port;
    } else {
      addr = file_node.addr;
    }
    fromclient = (file_node.addr.sin_addr.s_addr != addr.sin_addr.s_addr ||
		  file_node.addr.sin_port != addr.sin_port);

    if( list.size() > 0 ) {
      pthread_mutex_lock( &_nodes_in_use_lock );
      if( _nodes_in_use[fh.fid][addr.sin_addr.s_addr] > limit[i] ) {
	// if we're already fetching a block from this file from this nodes,
	// try an alternate if there are any
	status = wfs_protocol::IOERR;
	if( i < (list.size()-1) ) {
	  i++;
	} else {
	  i = 0;
	  limit[i]++;
	}
	pthread_mutex_unlock( &_nodes_in_use_lock );
	continue;
      }
      _nodes_in_use[fh.fid][addr.sin_addr.s_addr]++;
      pthread_mutex_unlock( &_nodes_in_use_lock );
    }

    if( wfs_dbg_block(WFS_DBG_1)) {
      if (file_node.addr.sin_addr.s_addr != addr.sin_addr.s_addr ||
          file_node.addr.sin_port != addr.sin_port) {
        wfs_log(WFS_DBG_1, "wfs_bt_client::readfile read cached client block from %s:%d fh %016"PRIx64" ver %d\n",
            inet_ntoa(addr.sin_addr), ntohs(addr.sin_port), fid,
            version.version_num);
      } else {
        wfs_log(WFS_DBG_1, "wfs_bt_client::readfile read from file node fh %016"PRIx64"\n", fid);
      }
    }

    status = wfs_client::readfile_from(addr, fh, version, off, n, user, info,
				       fromclient);
    buf.swap(info.buf); // avoid mem copy hopefully!

    if( list.size() > 0 ) {
      pthread_mutex_lock( &_nodes_in_use_lock );
      _nodes_in_use[fh.fid][addr.sin_addr.s_addr]--;
      if( _nodes_in_use[fh.fid][addr.sin_addr.s_addr] <= 0 ) {
	_nodes_in_use[fh.fid].erase(addr.sin_addr.s_addr);
	if( _nodes_in_use[fh.fid].size() > 0 ) {
	  _nodes_in_use.erase(fh.fid);
	}
      }
      pthread_mutex_unlock( &_nodes_in_use_lock );
    }

#ifdef WFS_WITH_LIBSSH
    if (status == wfs_protocol::OK) {
      // Check checksum
      if (rpc_is_secure() && fromclient) {
	int boff;
	assert (off % WFS_BLOCK_SIZE == 0);
	assert (n % WFS_BLOCK_SIZE == 0);
	for (boff = 0; boff < n/WFS_BLOCK_SIZE; boff++) {
	  unsigned gblock = off/WFS_BLOCK_SIZE+boff;
	  if (curr->checksums.count(gblock) == 0) {
	    if (wfs_dbg_block(WFS_DBG_1)) {
	      wfs_log(WFS_DBG_1, "No btchecksum found for boff %u of off=%llu, n=%d! (gblock #%u, %u given)\n",
	          boff, off, n, gblock,curr->checksums.size());
	      wfs_log(WFS_DBG_2, "Checksums we have: ");
	      std::map<unsigned,cacheloc_ent::checksum>::iterator it;
	      for (it = curr->checksums.begin(); it != curr->checksums.end(); it++) {
	        wfs_log(WFS_DBG_2_NOHEADER, "\t\t%u ",it->first);
	      }
	    }
	    status = wfs_protocol::IOERR;
	    break;
	  }
	  unsigned char sum[SHA256_DIGEST_LENGTH];
	  SHA256(((unsigned char *)buf.c_str())+boff*WFS_BLOCK_SIZE,
		 std::min(WFS_BLOCK_SIZE,buf.size()-boff*WFS_BLOCK_SIZE),
		 sum);
	  if (memcmp(sum,curr->checksums[gblock].data,
		     SHA256_DIGEST_LENGTH) != 0) {
	    wfs_log(WFS_DBG_1_NOHEADER, "btchecksum mismatch %llx vs %llx for version %d (real len=%u; starts with %llx)!\n",
		     *((unsigned long long *)sum),
		     *((unsigned long long *)curr->checksums[gblock].data),
		     version.version_num,buf.size(),
		     *((unsigned long long *)buf.c_str()));
	      /*char fn[42];
		snprintf(fn,42,"/tmp/%llx", *((unsigned long long *)sum));
		int fd = open(fn,O_WRONLY | O_CREAT | O_TRUNC, 0644);
		write(fd,buf.c_str(),buf.size());
		close(fd);*/
	    status = wfs_protocol::IOERR;
	    break;
	  } else {
	    wfs_log(WFS_DBG_2_NOHEADER, 
		    "btchecksum checks out for gblock #%u!\n",
		     gblock);
	  }
	}
      }
    }
#endif

    i++;

    // TODO: limit time to maxtime, or do things in parallel?
  } while( status != wfs_protocol::OK && i < list.size() );

  if( status == wfs_protocol::OK )  {
    if (_me.id != file_node.id) {

      // ~xavid: we can't add this information, because we can't trust
      //         the client's checksums.  In the future, we should
      //         have the storage node sign a checksum record that
      //         clients can pass to each other
#ifdef WFS_WITH_LIBSSH
      if (!rpc_is_secure()) {
#endif

      // save the block map that they gave us
      std::vector<cacheloc_ent> block_map_list;
      for( unsigned j = 0; j < info.block_map.size(); j++ ) {
	int32_t block = info.block_map[j];
	for( unsigned k = 0; k < sizeof(int32_t)*8; k++ ) {
	  if( block & 1 == 1 ) {
	    cacheloc_ent e;
	    e.s_addr = addr.sin_addr.s_addr;
	    e.s_port = addr.sin_port;
	    e.off.push_back((j*sizeof(int32_t)*8+k)*WFS_BLOCK_SIZE);
	    e.n.push_back(WFS_BLOCK_SIZE);
	    _viv->get_coords( addr.sin_addr.s_addr, e.c );
	    block_map_list.push_back(e);
	    wfs_log(WFS_DBG_1, "wfs_bt_client::readfile node %s:%d fh %016"PRIx64" reports cached block %u\n",
		     inet_ntoa(addr.sin_addr), ntohs(addr.sin_port), fid,
		     j*sizeof(int32_t)*8+k);
	  }
	  block >>= 1;
	}
      }
      _cache_loc.update_cacheloc( fid, version, block_map_list );

      if( wfs_dbg_block(WFS_DBG_1) && info.others.size() > 0 ) {
        wfs_log(WFS_DBG_1, "Also gained info about nodes: " );
        for( unsigned j = 0; j < info.others.size(); j++ ) {
          sockaddr_in addr_x;
          addr_x.sin_addr.s_addr = info.others[j].s_addr;
          wfs_log(WFS_DBG_1_NOHEADER, "\t\t%s:%d, ", inet_ntoa(addr_x.sin_addr),
              ntohs(info.others[j].s_port) );
        }
	wfs_log(WFS_DBG_1_NOHEADER, "\n");
      }
      _cache_loc.update_cacheloc( fid, version, info.others );
#ifdef WFS_WITH_LIBSSH
      }
#endif

      // let the server know we have it
      add_cache_struct acs;
      acs.file_node = file_node;
      acs.fh = fh;
      acs.version = version;
      acs.n = n;
      acs.off = off;

      pthread_mutex_lock( &_acs_lock );
      _acs_queue.push_back(acs);
      pthread_cond_signal( &_acs_condvar );
      pthread_mutex_unlock( &_acs_lock );

      // register the block in my block map
      pthread_mutex_lock( &_my_block_map_lock );
      std::vector<int32_t> map = _my_block_map[acs.fh.fid];
      unsigned int nb;
      unsigned int block_no = calc_block_no( acs.off, 1, nb );
      // each entry in the vector covers 32 blocks:
      unsigned int vec_index = (int) (block_no/(sizeof(int32_t)*8));
      while( map.size() < (vec_index+1) ) {
	map.push_back(0);
      }
      int32_t mask = 1;
      mask <<= (block_no - (vec_index*sizeof(int32_t)*8));
      map[vec_index] = map[vec_index] | mask;
      wfs_log(WFS_DBG_1, "Updating block mask for file %016"PRIx64" block %d vec_index %d mask %d\n",
          acs.fh.fid, block_no, vec_index, mask );
      _my_block_map[acs.fh.fid] = map;
      pthread_mutex_unlock( &_my_block_map_lock );

    }
  }
  return status;

}

void wfs_bt_client::add_cache_loc() {

  while( true ) {

    pthread_mutex_lock( &_acs_lock );
    while( _acs_queue.size() == 0 ) {
      pthread_cond_wait( &_acs_condvar, &_acs_lock );
    }

    // now tell the responsible node about me
    cacheloc_ent e;
    e.s_addr = _bt_addr.sin_addr.s_addr;
    e.s_port = _bt_addr.sin_port;
    e.c = _viv->my_coords();

    std::vector<add_cache_struct> new_acs_queue;
    wfs_node file_node = _acs_queue.front().file_node;
    wfs_file_handle fh = _acs_queue.front().fh;
    wfs_version version = _acs_queue.front().version;
    for( unsigned i = 0; i < _acs_queue.size(); i++ ) {
      add_cache_struct acs = _acs_queue[i];
      if( acs.file_node.id == file_node.id &&
	  acs.fh == fh ) {
	e.off.push_back(acs.off);
	e.n.push_back(acs.n);
      } else {
	new_acs_queue.push_back(acs);
      }
    }
    _acs_queue.clear();
    _acs_queue = new_acs_queue;

    wfs_log(WFS_DBG_1, "pid %d finish readfile update cacheloc at file_node %s:%d fid %016"PRIx64" off %llu %llu, my coords are %s\n",
	     getpid(), inet_ntoa(file_node.addr.sin_addr),
	     ntohs(file_node.addr.sin_port), file_handle_to_file_id(fh),
	     e.off[0], e.n[0],
	     e.c.print_coord().c_str());
    int dummy;
    pthread_mutex_unlock( &_acs_lock );
    wfs_protocol::status ss;
    wfs_log(WFS_DBG_3, "get_client temp "__FILE__".%d: %s\n",
	    __LINE__, inet_ntoa(file_node.addr.sin_addr));
    rpcc *client = _clients->get_client( file_node, fhto(fh) );
    if( client == NULL ) {
      ss = rpc_const::timeout_failure;
    } else {
      ss = client->call( wfs_protocol::addcacheloc, fh, version, e, dummy,
			 fhto(fh), "host");
      _clients->done_with_client(file_node);
    }
    if (ss != wfs_protocol::OK) {
      wfs_log(WFS_DBG_1, "Got a %d from addcacheloc!\n", ss);
    }
  }

}
