/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * overfuse creates a FUSE file system and sends the file-system
 * operations to overserve via the wfs_client RPC interface.
 *
 * as root, mkdir /mnt/rtm ; chown rtm /mnt/rtm
 * as rtm, ./overserve &
 * as rtm, ./overfuse /mnt/rtm
 *
 * started life as examples/overfuse.c in the fuse distribution.
 * we have to use low-level interface in order to get i-numbers.
 * the high-level interface only gives us complete paths.
 */

#include "wfs_cache_client.h"
#include "wfs_bt_client.h"
#include "wfs_options.h"
#include <arpa/inet.h>
#include <assert.h>
#include <chash.h>
#include <errno.h>
#include <fcntl.h>
#include <fuse_lowlevel.h>
#include <ext/hash_map>
#include <map>
#include <vector>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <slice_manager.h>
#include <wfs_cache_mem.h>
#include <wfs_misc.h>
#include <sys/resource.h>
#include <sys/time.h>

//#define GOOGLE_HEAP 1

#ifdef GOOGLE_HEAP
#include <google/heap-profiler.h>
#endif

using __gnu_cxx::hash_map;

wfs_client *odc;

std::map< std::string, std::vector<int> > stats;

pthread_mutex_t stats_lock;

//#define WFS_STATS 1

#ifdef WFS_STATS

inline float start_stat() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return (tv.tv_sec*1000 + ((float) tv.tv_usec/1000));
}

inline void record_stat( std::string name, float start ) {
  float end = start_stat();
  float time = end - start;
  int bin;
  if( time < 1 ) {
    bin = 0;
  } else if( time < 5 ) {
    bin = 1;
  } else if( time < 10 ) {
    bin = 2;
  } else if( time < 20 ) {
    bin = 3;
  } else if( time < 50 ) {
    bin = 4;
  } else if( time < 100 ) {
    bin = 5;
  } else if( time < 200 ) {
    bin = 6;
  } else if( time < 500 ) {
    bin = 7;
  } else {
    bin = 8;
  }
  pthread_mutex_lock( &stats_lock );
  if( stats[name].size() <= bin ) {
    stats[name].resize(bin+1, 0);
  }
  stats[name][bin]++;
  pthread_mutex_unlock( &stats_lock );
}

void print_stats( int x ) {
  std::map< std::string, std::vector<int> >::const_iterator i = stats.begin();
  for( ; i != stats.end(); i++ ) {
    printf( "%s: ", i->first.c_str() );
    for( int j = 0; j < i->second.size(); j++ ) {
      printf( "%d ", i->second[j] );
    }
    printf( "\n" );
  }
  exit(0);
}

#else

inline float start_stat() { return 0; }
inline void record_stat( std::string name, float start ) {}
void print_stats( int x ) { exit(0); }

#endif

inline int errno_from_status(wfs_protocol::status s) {
  switch (s) {
  case wfs_protocol::NOENT:
    return ENOENT;
  case wfs_protocol::IOERR:
    return EIO;
  case wfs_protocol::EXIST:
    return EEXIST;
  case wfs_protocol::NOTDIR:
    return ENOTDIR;
  case wfs_protocol::ISDIR:
    return EISDIR;
  case wfs_protocol::NOTEMPTY:
    return ENOTEMPTY;
  case wfs_protocol::PERM:
    return EACCES;
  case wfs_protocol::STALE:
    return ESTALE;
  case wfs_protocol::RPCERR:
    return EBADMSG;
#ifdef EREMOTE
  case wfs_protocol::NOTMASTER:
    return EREMOTE;
#endif
#ifdef EBADRQC
  case wfs_protocol::BADVERSION:
    return EBADRQC;
#endif
  case rpc_const::timeout_failure:
    return ETIMEDOUT;
  default:
    wfs_log(WFS_DBG_1, "errno_from_status: Got a status %d!\n",(int)s);
    return EIO;
  }
}

const long GR_MAX_NAME = sysconf(_SC_GETGR_R_SIZE_MAX);
hash_map<gid_t, std::string> grnam_cache;
std::map<std::string, gid_t> grgid_cache;
inline std::string get_grnam_from_gid( gid_t gid ) {

  if( grnam_cache.find(gid) != grnam_cache.end() ) {
    return grnam_cache[gid];
  }

  struct group g;
  char name[GR_MAX_NAME];
  struct group *res = NULL;
  if (gid == (gid_t)-1)
    return "";
  else if( getgrgid_r( gid, &g, name, GR_MAX_NAME, &res ) || res == NULL) {
    wfs_log(WFS_DBG_1, "getgrgid failed for group %d\n", gid );
    return "";
  }
  std::string s = std::string(g.gr_name);
  grnam_cache[gid] = s;
  grgid_cache[s] = gid;
  return std::string(g.gr_name);
}

bool no_permissions;

inline std::string req_ident(const fuse_req_t &req) {
  if( no_permissions ) {
    return ROOT_IDENT;
  }
  uid_t u = fuse_req_ctx(req)->uid;
  if (u == getuid() && odc->has_local_server()) {
    return ROOT_IDENT;
  } else
    return get_pwnam_from_uid(u);
}

typedef struct wfs_fuse_ino {
  wfs_fuse_ino( wfs_file_handle new_fh ) :
      fh(new_fh), version(0,0), parent(0) {}
  wfs_file_handle fh;
  wfs_version version;
  fuse_ino_t parent;
} wfs_file_ino;

typedef struct fh_to_ino_entry {
  wfs_fuse_ino *ino;
  fh_to_ino_entry *next;
} fh_to_ino_entry;

pthread_mutex_t fh_to_ino_lock;
hash_map<wfs_file_id, fh_to_ino_entry *> fh_to_ino;

std::map<std::string, uid_t> pwuid_cache;

/* xavid: These convert between WFS permissions and local mode/uid/gid */
void fill_stat_for_permissions(struct stat &st, const std::map<std::string,
			       wfs_permission> &permissions,
			       const std::string &explicit_ident = "") {
  /* A WFS file can have permissions for many users and groups.  We can only
     show one of each in our default stat struct.  Ideally, this shouldn't
     depend on the current user, to avoid confusing people who sudo to
     chown/chmod.  Note that we enforce permissions, not FUSE, so what we
     show here should be whatever is most useful to the user.

     Strategy: the owner is the first std::string ASCIIbetically* who
     has the admin bit and has a passwd entry.  If no such std::string exists,
     say the owner is uid 0.

     The group is the first std::string ASCIIbetically who has a different
     name than the owner and has a group entry.  If no such group exists,
     the group is gid 0.

     Note: this function leaves the type bits of st.st_mode blank.

     If an explicit .std::string=bob cue is given, show bob's permissions for
     the user and group entries.  Still show ANYONE_IDENT for other.
     If the explicit std::string has WFS_PERMISSION_ADMIN and has a uid, show
     him as the owner, otherwise show uid 0.  If he has a gid, show
     him as the group, otherwise show gid 0.

     *: Conveniently, std::map iterators return pairs in ascending order
        by key.
  */

  st.st_mode = 0;
  st.st_uid = 0;
  st.st_gid = 0;

  // Find the owner
  std::map<std::string, wfs_permission>::const_iterator i;
  std::string uname;

  wfs_permission baseline;
  i = permissions.find(ANYONE_IDENT);
  if (i != permissions.end())
    baseline = i->second;
  else
    baseline = 0;

  struct passwd pwent;
  char name[PW_MAX_NAME];
  struct passwd *res = NULL;

  uid_t pwgood = 0;

  if (explicit_ident.length() > 0) {
    i = permissions.find(explicit_ident);
    if (i != permissions.end()) {

      if( pwuid_cache.find(i->first) != pwuid_cache.end() ) {
	pwgood = pwuid_cache[i->first];
      } else if (getpwnam_r( i->first.c_str(), &pwent, name, PW_MAX_NAME, &res )
	  || res == NULL) {
        wfs_log(WFS_DBG_2, "getpwnam failed for explicit std::string %s\n", i->first.c_str() );
      } else {
	pwgood = pwent.pw_uid;
	pwuid_cache[i->first] = pwgood;
      }
    }
  } else {
    for (i = permissions.begin(); i != permissions.end(); i++) {
      if (i->second & WFS_PERMISSION_ADMIN
	  && i->first != ANYONE_IDENT) {
	if( pwuid_cache.find(i->first) != pwuid_cache.end() ) {
	  pwgood = pwuid_cache[i->first];
	  break;
	} else if( getpwnam_r( i->first.c_str(), &pwent, name,
			       PW_MAX_NAME, &res )
	    || res == NULL) {
	  wfs_log(WFS_DBG_2, "getpwnam failed for name %s\n", i->first.c_str() );
	} else {
	  pwgood = pwent.pw_uid;
	  pwuid_cache[i->first] = pwgood;
	  break;
	}
      }
    }
  }
  if (i == permissions.end()) {
    if (explicit_ident.length() > 0) {
      // Leave at 0
    } else {
      // Root has everything all the time
      st.st_mode |= S_IRUSR | S_IWUSR;
      // except maybe not execute, for compatability reasons
      for (i = permissions.begin(); i != permissions.end(); i++) {
	if (i->second & WFS_PERMISSION_EXECUTE) {
	  st.st_mode |= S_IXUSR;
	  break;
	}
      }
    }
  } else {
    uname = i->first;
    if (i->second & WFS_PERMISSION_ADMIN) // Might not with an explicit .std::string=
      st.st_uid = pwgood;
    if (i->second & WFS_PERMISSION_READ)
      st.st_mode |= S_IRUSR;
    if (i->second & WFS_PERMISSION_WRITE)
      st.st_mode |= S_IWUSR;
    if (i->second & WFS_PERMISSION_EXECUTE) {
      st.st_mode |= S_IXUSR;
    }
  }
  struct group grent;
  char gname[GR_MAX_NAME];
  grent.gr_gid = 0;
  struct group *gres = NULL;
  if (explicit_ident.length() > 0) {
    //i = permissions.find(explicit_ident); -- i still is this
    if(i != permissions.end() ) {
      if( grgid_cache.find(i->first) != grgid_cache.end() ) {
        grent.gr_gid = grgid_cache[i->first];
      } else if(getgrnam_r( i->first.c_str(), &grent, gname,
          GR_MAX_NAME, &gres) || gres == NULL) {
        wfs_log(WFS_DBG_2, "getgrnam failed for explicit std::string %s\n",
            i->first.c_str() );
      } else {
        grgid_cache[i->first] = grent.gr_gid;
        grnam_cache[grent.gr_gid] = i->first;
      }
    }
  } else {
    for (i = permissions.begin(); i != permissions.end(); i++) {
      if (i->first != uname && i->first != ANYONE_IDENT && i->second > 0) {

	if( grgid_cache.find(i->first) != grgid_cache.end() ) {
	   grent.gr_gid = grgid_cache[i->first];
	   break;
	} else if( getgrnam_r( i->first.c_str(), &grent, gname,
			       GR_MAX_NAME, &gres)
		   || gres == NULL) {
	  wfs_log(WFS_DBG_2, "getgrnam failed for name %s\n", i->first.c_str() );
	} else {
	  grgid_cache[i->first] = grent.gr_gid;
	  grnam_cache[grent.gr_gid] = i->first;
	  break;
	}
      }
    }
  }

  if (i == permissions.end()) {
    // Leave it as 0
  } else {
    if (i->second & WFS_PERMISSION_READ)
      st.st_mode |= S_IRGRP;
    if (i->second & WFS_PERMISSION_WRITE)
      st.st_mode |= S_IWGRP;
    if (i->second & WFS_PERMISSION_EXECUTE)
      st.st_mode |= S_IXGRP;
  }
  // This will either properly keep it at 0, or set it to the right user
  st.st_gid = grent.gr_gid;

  if (baseline & WFS_PERMISSION_READ)
    st.st_mode |= S_IROTH;
  if (baseline & WFS_PERMISSION_WRITE)
    st.st_mode |= S_IWOTH;
  if (baseline & WFS_PERMISSION_EXECUTE)
    st.st_mode |= S_IXOTH;

  wfs_log(WFS_DBG_1, "Mode for perms %s is %#o. (ei='%s')\n",
	   format_permissions(permissions).c_str(),
	   st.st_mode, explicit_ident.c_str());
}

/* This function is intended to lead to the following behavior:

   Nota bene: since chown on many systems can only be run by root, if
              chown is the only way of changing a file's ADMIN perm there's
	      no way to "give files away" unless you're root.  This may
	      or may not be a problem.

   (current owner and group are as reported by ls -l,
    except that if uid 0 is reported and some nonlocal
    user has ADMIN, that user is treated as the owner, and if gid 0
    is reported and some user other than the owner has permissions,
    that user is treated as the group)

   Without .std::string=: (these assume there exists a local user bob)
   chown bob file: bob gets all permissions of current owner, current owner
                   loses those permissions; other std::strings unaffected
   chgrp bob file: bob gets all permissions of current group except ADMIN;
                   if current group had ADMIN, it keeps all permissions,
		   otherwise it loses all permissions
   chmod $m file: affects only the permissions reported by ls -l

   With .std::string=: (no need for a local user bob; anyone is any local user)
   chown anyone .std::string=bob/file: bob gets the old owner's permissions,
                                 including ADMIN; old owner loses all
				 permissions; other std::strings unaffected
   chgrp anygroup .std::string=bob/file: ???
   chmod $m .std::string=bob/file: sets the permissions of bob based on the
                             the specified user permissions;
			     sets anyone based on the other permissions;
			     does not change bob's ADMIN perm or any other
			     permissions
 */
std::map<std::string, wfs_permission> permissions_for_mode(mode_t mode,
					      uid_t user,
					      gid_t group,
					      wfs_file_handle &fh,
					      fuse_req_t &req,
					      bool new_file) {

  std::string explicit_ident = (fh.which_cues&WFS_CUE_IDENT
			      ?fh.cues[wfs_file_handle::CUE_SLOT_IDENT]
			      .strval
			      :"");

  std::map<std::string, wfs_permission> perms;
  std::string old_owner, old_group;
  // These are true if the old_owner or old_group don't match ls -l
  bool ubottomed = false, gbottomed = false;

  /* The old group and user are determined as for fill_stat_for_permissions;
     however, if that would use uid or gid 0, this instead uses the first
     nonroot std::string with ADMIN and the first other nonroot std::string, respectively.
  */

  // Keep this function's behavior in sync with fill_stat_for_permissions's
  // (except it does a second pass if the first pass fails, for nonlocal users)
  // -? Can they be unified?
  old_owner = "";
  old_group = "";

  std::map<std::string,wfs_permission> old_perms;
  if (!new_file) {
    wfs_protocol::status ret;

    if (odc->isfile(fh)) {
      wfs_protocol::fileinfo info;
      ret = odc->getfile(fh, req_ident(req), info);
      if( ret != wfs_protocol::OK ) {
        wfs_log(WFS_DBG_1, "get_owner_group %016"PRIx64" bad file get %d\n",
            fh.fid, ret );
        return perms;
      }
      old_perms = info.permissions;
    } else { // dir
      wfs_protocol::dirinfo info;
      ret = odc->getdir(fh, req_ident(req), info);
      if( ret != wfs_protocol::OK ) {
        wfs_log(WFS_DBG_1, "get_owner_group %016"PRIx64" bad dir get %d\n",
            fh.fid, ret );
        return perms;
      }
      old_perms = info.permissions;
    }

    std::map<std::string, wfs_permission>::const_iterator i;

    struct passwd pwent;
    char name[PW_MAX_NAME];
    struct passwd *res = NULL;

    for (i = old_perms.begin(); i != old_perms.end(); i++) {
      if (i->second & WFS_PERMISSION_ADMIN && i->first != ANYONE_IDENT) {
        if( getpwnam_r( i->first.c_str(), &pwent, name, PW_MAX_NAME, &res )
            || res == NULL) {
          wfs_log(WFS_DBG_2, "gog: getpwnam failed for name %s\n", i->first.c_str() );
        } else {
          break;
        }
      }
    }
    if (i == old_perms.end()) {
      ubottomed = true;
      for (i = old_perms.begin(); i != old_perms.end(); i++) {
        if (i->second & WFS_PERMISSION_ADMIN && i->first != ANYONE_IDENT) {
          break;
        }
      }
    }
    if (i != old_perms.end())
      old_owner = i->first;

    wfs_permission baseline;
    i = old_perms.find(ANYONE_IDENT);
    if (i != old_perms.end())
      baseline = i->second;
    else
      baseline = 0;

    struct group grent;
    char gname[GR_MAX_NAME];
    struct group *gres = NULL;

    // Now, look for group
    for (i = old_perms.begin(); i != old_perms.end(); i++) {
      if (i->first != old_owner && i->first != ANYONE_IDENT && i->second > 0) {
        if( getgrnam_r( i->first.c_str(), &grent, gname, GR_MAX_NAME, &gres)
            || gres == NULL) {
          wfs_log(WFS_DBG_2, "gog: getgrnam failed for name %s\n", i->first.c_str() );
        } else {
          break;
        }
      }
    }
    if (i == old_perms.end()) {
      gbottomed = true;
      for (i = old_perms.begin(); i != old_perms.end(); i++) {
        if (i->first != old_owner && i->first != ANYONE_IDENT
            && i->second > 0) {
          break;
        }
      }
    }
    if (i != old_perms.end())
      old_group = i->first;
  }
  // End get_owner_group

  if (user != (uid_t)-1 && !new_file) {
    if (old_owner.length() > 0) {
      perms[old_owner] = 0;
      wfs_log(WFS_DBG_2, "permissions_for_mode: clearing permissions for user %s\n",
          old_owner.c_str());
    }
  }

  if (explicit_ident.length() > 0) {
    if (mode == (mode_t)-1) {
      perms[explicit_ident] = old_perms[old_owner];
    } else {
      perms[explicit_ident] =
        (((user != (uid_t)-1 || new_file) ? WFS_PERMISSION_ADMIN :0)
            | ((mode & S_IRUSR) ? WFS_PERMISSION_READ :0)
            | ((mode & S_IWUSR) ? WFS_PERMISSION_WRITE :0)
            | ((mode & S_IXUSR) ? WFS_PERMISSION_EXECUTE :0));
    }
    wfs_log(WFS_DBG_2, "permissions_for_mode: %s explicitly has %x\n",
        explicit_ident.c_str(),
        perms[explicit_ident]);
  } else { // No explicit std::string
    if (group != (gid_t)-1 && !new_file) {
      if (old_group != "") {
        perms[old_group] = 0;
        wfs_log(WFS_DBG_2, "permissions_for_mode: clearing permissions for group %s\n",
            old_group.c_str());
      }
    }

    // Do these after the old, because each time they're called they reuse memory
    std::string uinfo_name, ginfo_name;
    if (user != (uid_t)-1)
      uinfo_name = get_pwnam_from_uid(user);
    else
      uinfo_name = ubottomed?"":old_owner;
    if (new_file)
      ginfo_name = ""; // Don't use the creator's primary group for anything
    else if (group != (gid_t)-1)
      ginfo_name = get_grnam_from_gid(group);
    else
      ginfo_name = gbottomed?"":old_group;

    if (uinfo_name != "" && uinfo_name != "nobody" ) {
      if (mode == (mode_t)-1) {
        perms[uinfo_name] = old_perms[old_owner];
      } else {
        perms[uinfo_name] = (WFS_PERMISSION_ADMIN
            | ((mode & S_IRUSR) ? WFS_PERMISSION_READ :0)
            | ((mode & S_IWUSR) ? WFS_PERMISSION_WRITE :0)
            | ((mode & S_IXUSR) ? WFS_PERMISSION_EXECUTE :0));
      }
      wfs_log(WFS_DBG_2, "permissions_for_mode: %s has %x\n", uinfo_name.c_str(),
          perms[uinfo_name]);
    }

    if (ginfo_name != ""
      && ginfo_name != "nobody"
        && ginfo_name != "nogroup"
          && (uinfo_name == "" || ginfo_name != uinfo_name)) {
      if (mode == (mode_t)-1) {
        perms[ginfo_name] = old_perms[old_group];
      } else {
        perms[ginfo_name] = (((mode & S_IRGRP) ? WFS_PERMISSION_READ :0)
            | ((mode & S_IWGRP) ? WFS_PERMISSION_WRITE :0)
            | ((mode & S_IXGRP) ? WFS_PERMISSION_EXECUTE :0));
      }
      wfs_log(WFS_DBG_2, "permissions_for_mode: %s has %x\n",ginfo_name.c_str(),
          perms[ginfo_name]);
    }
  }
  if (mode == (mode_t)-1) {
    perms[ANYONE_IDENT] = old_perms[ANYONE_IDENT];
  } else {
    perms[ANYONE_IDENT] = (((mode & S_IROTH) ? WFS_PERMISSION_READ :0)
			   | ((mode & S_IWOTH) ? WFS_PERMISSION_WRITE :0)
			   | ((mode & S_IXOTH) ? WFS_PERMISSION_EXECUTE :0));
  }
  wfs_log(WFS_DBG_2, "permissions_for_mode: %s has %x\n",ANYONE_IDENT.c_str(),
	   perms[ANYONE_IDENT]);

  return perms;
}


/***
 * The following variables and methods are an attempt to fix a
 * multithreading problem with the FreeBSD FUSE implementation.  If
 * fuse4bsd starts a create operation, and the concurrently starts a
 * lookup operation, WFS will (correctly) return the same ino for both
 * operatuions.  However, if the result of the lookup gets processed
 * first, fuse4bsd will put the inumber in the vfs hash table.  Then,
 * when the fuse4bsd code processes the result of the create
 * operation, it will check the vfs hash table and find a collision.
 * This causes it to issue a FORGET operation on the ino, which in
 * turn causes subsequent operation on that file handle by the
 * application to do operations on the zero file handle, causing many
 * problems.
 *
 * My solution is to keep creates and lookups from happening
 * concurrently.  This can probably be made more efficient by only
 * keeping creates and lookups on the same dir/name from happening
 * concurrently, but I didn't feel like managing different condition
 * variables, as that would be a giant pain.
 *
 * However!  Even though this code keeps the operations from happening
 * concurrently within WheelFS, it does not absolutely solve the
 * issue, since fuse4bsd can still be scheduled to process the
 * requests in the error-causing order.  But hopefully this makes this
 * rare problem happen even more rarely.
 *
 * Not sure if this issue also exists on Linux; I've never seen it happen.
 * Leaving it disabled for now, so it doesn't impact multithreaded
 * performance.
 *
 * ?xavid: This currently also happens on Mac because of the FreeBSD hack;
 *         not sure if it's necessary or not.
 *
 */
pthread_mutex_t create_lock;
bool creating_or_lookuping = false;
pthread_cond_t create_condvar;

inline void start_c_or_l() {
#ifdef __FreeBSD__
  pthread_mutex_lock( &create_lock );
  while( creating_or_lookuping ) {
    wfs_log(WFS_DBG_1, "Waiting for creating/lookuping lock\n" );
    pthread_cond_wait( &create_condvar, &create_lock );
  }
  creating_or_lookuping = true;
  pthread_mutex_unlock( &create_lock );
#endif
}

inline void end_c_or_l() {
#ifdef __FreeBSD__
  pthread_mutex_lock( &create_lock );
  creating_or_lookuping = false;
  pthread_cond_broadcast( &create_condvar );
  pthread_mutex_unlock( &create_lock );
#endif

  // fix highly terrible scheduling bug when using multithreaded FUSE on Linux.
  // Without this, the client can consume 100% of the CPU when serving items
  // out of its cache, and cause starvation problems for other pthreads in the
  // wfs_server process.  Probably deserves to go at the end of every operation
  // handler, but it should be sufficient to do after every lookup instead . . .
#ifndef __APPLE__
  pthread_yield();
#else
  // pthread_yield isn't available on Mac
  pthread_yield_np();
#endif
}

/***
 * END terrible fuse4bsd-bug-related fixes
 */

wfs_file_handle get_fh( fuse_ino_t ino ) {
  // fuse insists that the root directory has an ino of 1
  wfs_file_handle fh;
  if( ino == 1 ) {
    fh.fid = 1;
  } else {
    fh = ((wfs_fuse_ino*) ino)->fh;
  }
  return fh;
}

inline void set_thread_id( wfs_file_handle &fh, fuse_req_t req ) {

  if( wfs_dbg_block(2) ) {
    const struct fuse_ctx *ctx = fuse_req_ctx( req );
    int32_t id = wfs_set_thread_id( (int32_t) ctx->pid );
    fh.set_cue( WFS_CUE_ID, id );
  }

}

inline void clear_thread_id() {

  if( wfs_dbg_block(2) ) {
    wfs_set_thread_id(0);
  }

}

wfs_version get_ver( fuse_ino_t ino ) {
  // fuse insists that the root directory has an ino of 1
  if( ino == 1 ) {
    return wfs_version();
  } else {
    wfs_version retv = ((wfs_fuse_ino*) ino)->version;
    // Handle explicit version cue
    // xavid: I'm not sure I like putting this in the fuse client, but
    //        it does need to get in before or in the cache client, at least
    wfs_cue vcue = ((wfs_fuse_ino*)ino)->fh.get_cue(WFS_CUE_VERSION_PARAM);
    if (vcue.val > 0) {
      wfs_log(WFS_DBG_1, "overriding version=%d to %d for .version= cue\n",
          retv.version_num, vcue.val);
      retv.version_num = vcue.val;
    }
    return retv;
  }
}

void set_ver( fuse_ino_t ino, wfs_version version ) {
  // fuse insists that the root directory has an ino of 1
  if( ino == 1 ) {
    return;
  } else {
    wfs_log(WFS_DBG_1, "Updating version of %016"PRIx64" from %d to %d\n",
	      get_fh(ino).fid,
	      ((wfs_fuse_ino*) ino)->version.version_num,
	      version.version_num);
    ((wfs_fuse_ino*) ino)->version = version;
  }
}

// assume fh_to_ino_lock is already held
fh_to_ino_entry *get_fh_to_ino_entry( wfs_file_handle fh,
				      fuse_ino_t parent = 0 ) {

  fh_to_ino_entry *entry = fh_to_ino[fh.fid];
  while( entry != NULL ) {
    // find the entry that matches this cue set
    // we have to consider the parent here because if we create
    // a file under a directory that has permanent cues, and then
    // later read that file under the same directory without
    // any cues, the file's handle will look the same as far as cues
    // go.  Without considering the parent, we will return the same
    // ino, which confuses the file system because it looks like
    // the same ino was assigned to two different files.
    if( fh == entry->ino->fh &&
	(parent == 0 || parent == entry->ino->parent) ) {
      break;
    } else {
      entry = entry->next;
    }
  }
  return entry;

}

fuse_ino_t get_fuse_ino( wfs_file_handle fh, fuse_ino_t parent = 0 ) {

  // Note that the data pointed to by these inode numbers will never
  // get deleted until this client exits, because there's no way of
  // knowing whether FUSE will ever use them or not once we've created
  // them and passed them back.  That's why it's important to reuse them
  // whenever possible.
  pthread_mutex_lock( &fh_to_ino_lock );

  fh_to_ino_entry *entry = get_fh_to_ino_entry(fh, parent);

  if( entry != NULL ) {
    pthread_mutex_unlock( &fh_to_ino_lock );
    return (fuse_ino_t) entry->ino;
  }

  // have to make a new one now
  fh_to_ino_entry *new_entry = new fh_to_ino_entry();
  new_entry->ino = new wfs_fuse_ino( fh );
  new_entry->next = fh_to_ino[fh.fid];
  new_entry->ino->parent = parent;
  fh_to_ino[fh.fid] = new_entry;

  wfs_log(WFS_DBG_1, "New ino for file handle %s, p=%d == %ld\n",
	    fh.print().c_str(), (int)parent, (long) new_entry->ino );

  // set the version of this guy to the max version of any saved inode
  fh_to_ino_entry *next = new_entry->next;
  while( next != NULL ) {
    if( new_entry->ino->version < next->ino->version ) {
      set_ver( (fuse_ino_t) new_entry->ino, next->ino->version );
    }
    next = next->next;
  }
  pthread_mutex_unlock( &fh_to_ino_lock );
  return (fuse_ino_t) new_entry->ino;
}

wfs_protocol::status getattr(fuse_req_t &req,
			     wfs_file_handle &fh, struct stat &st,
			     bool same_as_parent = false,
			     bool ignore_ident = false) {

  wfs_protocol::status ret;
  bzero( &st, sizeof(st) );

  wfs_log(WFS_DBG_1, "wfs_fuse_client::getattr %016"PRIx64"\n", fh.fid );

  // because the permanent cues are stored on the file server
  // (and not the dir server from where we got the original file handle)
  // we might need to update the permanent cues here
  pthread_mutex_lock( &fh_to_ino_lock );
  fh_to_ino_entry *entry = get_fh_to_ino_entry(fh);
  pthread_mutex_unlock( &fh_to_ino_lock );

  if( odc->isfile(fh) ) {

    wfs_protocol::fileinfo info;
    ret = odc->getfile(fh, req_ident(req), info);
    if( ret != wfs_protocol::OK ) {
      wfs_log(WFS_DBG_1, "wfs_fuse_client::getattr %016"PRIx64" bad file get %d\n",
          fh.fid, ret );
      return ret;
    } else {
      wfs_log(WFS_DBG_1, "Got finfo for file %016"PRIx64" with size %lld\n",
	      fh.fid, info.size );
    }

    if( entry != NULL ) {
      set_ver( (fuse_ino_t) entry->ino, info.version );
      st.st_ino = (fuse_ino_t) entry->ino;
    } else {
      fuse_ino_t ino = get_fuse_ino(fh);
      assert( ino );
      set_ver( ino, info.version );
      st.st_ino = ino;
    }

    fill_stat_for_permissions(st, info.permissions,
			      ((!ignore_ident && fh.which_cues&WFS_CUE_IDENT)
			       ?fh.cues[wfs_file_handle::CUE_SLOT_IDENT].strval
			       :""));
    if( odc->islink(fh) ) {
      st.st_mode |= S_IFLNK;
    } else {
      st.st_mode |= S_IFREG;
    }
    st.st_nlink = 1;
    st.st_size = info.size;
    st.st_atime = info.atime;
    st.st_mtime = info.mtime;
    st.st_ctime = info.ctime;

    wfs_log(WFS_DBG_1, "Got finfo for file %016"PRIx64" with size %ld\n",
	      fh.fid, (long)st.st_size );

  } else {

    // for dirs, if this fid is the same as its parents, then all we
    // did was set a cue.  no need to recheck with the directory server

    if( !same_as_parent ) {

      wfs_protocol::dirinfo info;
      ret = odc->getdir(fh, req_ident(req), info);
      if( ret != wfs_protocol::OK ) {
        wfs_log(WFS_DBG_1, "wfs_fuse_client::getattr %016"PRIx64" bad dir get %d\n",
            fh.fid, ret );
        return ret;
      }
      st.st_ino = get_fuse_ino(fh);
      fill_stat_for_permissions(st, info.permissions,
				(fh.which_cues&WFS_CUE_IDENT
				 ?fh.cues[wfs_file_handle::CUE_SLOT_IDENT]
				 .strval
				 :""));
      st.st_mode |= S_IFDIR;
      st.st_size = info.size;
      st.st_atime = info.atime;
      st.st_mtime = info.mtime;
      st.st_ctime = info.ctime;
      st.st_blocks = ((int) (info.size/512))+1;
    } else {
      st.st_mode = S_IFDIR | 0777;
      // TODO: correctly setting the times and mode?
    }

    // since it's hard to know the actual number of links in this directory
    // (should be 2 + the number of entries), we just set it to 1.  I believe
    // this is ok, and allows tools like 'find' to work correctly.
    st.st_nlink = 1;

  }

  // potentially set permanent cues here
  pthread_mutex_lock( &fh_to_ino_lock );
  if( entry != NULL && entry != get_fh_to_ino_entry(fh) ) {
    entry->ino->fh = fh;
  }
  pthread_mutex_unlock( &fh_to_ino_lock );

  return wfs_protocol::OK;

}


void wfs_fuse_client_getattr( fuse_req_t req, fuse_ino_t ino,
			      struct fuse_file_info *fi ) {

  float start = start_stat();
  std::string statname = "getattr";

  struct stat st;
  wfs_file_handle fh = get_fh(ino);
  wfs_protocol::status ret;

  set_thread_id( fh, req );

  ret = getattr(req, fh, st);
  if( ret != wfs_protocol::OK ) {
    fuse_reply_err( req, errno_from_status(ret) );
    record_stat( statname, start );
    return;
  }

  wfs_log(WFS_DBG_1, "Getattr returning for file %016"PRIx64" with size %ld\n",
	    fh.fid, (long)st.st_size );
  fuse_reply_attr( req, &st, 0.0 );
  clear_thread_id();
  record_stat( statname, start );
}

int create_helper( fuse_req_t &req, fuse_ino_t parent, const char *name,
		   mode_t mode, struct fuse_entry_param *e ) {

  wfs_file_handle pfh = get_fh(parent);
  set_thread_id( pfh, req );

  wfs_file_handle xfh;
  int status = 0;
  struct stat st;
  wfs_log(WFS_DBG_1, "wfs_fuse_client_create %016"PRIx64" %s mode %o\n",
	    pfh.fid, name, mode );

  const struct fuse_ctx *ctx = fuse_req_ctx(req);
  wfs_protocol::status ret = odc->create( pfh, name,
					  permissions_for_mode(mode,
							       ctx->uid,
							       ctx->gid,
							       pfh,req,true),
					  WFS_TYPE_REGULAR,
					  req_ident(req),
					  xfh );

  if( (ret == wfs_protocol::OK &&
       getattr( req, xfh, st ) == wfs_protocol::OK) ||
      (ret == wfs_protocol::EXIST &&
       odc->lookup( pfh, name, req_ident(req), xfh ) == wfs_protocol::OK &&
       getattr( req, xfh, st ) == wfs_protocol::OK) ) {

    bzero( e, sizeof(*e) );
    e->ino = get_fuse_ino( xfh, parent );
    assert( e->ino );
    e->attr_timeout = 0.0;
    e->entry_timeout = 0.0;
    e->attr = st;
    wfs_log(WFS_DBG_1, "Finishing create of file %s, new fid is %016"PRIx64" %d %d\n", name,
	       xfh.fid, (int) e->ino, (int) st.st_ino );
  } else {
    status = ENOENT;
  }

  clear_thread_id();
  return status;

}

void wfs_fuse_client_create( fuse_req_t req, fuse_ino_t parent,
			     const char *name, mode_t mode,
			     struct fuse_file_info *fi ) {

  float start = start_stat();
  std::string statname = "create";

  struct fuse_entry_param e;
  start_c_or_l();
  int ret = create_helper( req, parent, name, mode, &e );

  // xavid: We define __FreeBSD__ on OS X to work around stuff.
  // MacFUSE doesn't like direct_io.
#if defined(__FreeBSD__) && !defined(__APPLE__)
  // sigh.  FreeBSD's FUSE cache has bad bugs, but FUSE on Linux doesn't do
  // appends correctly when direct_io is enabled.
  if( !(mode & S_IXUSR || mode & S_IXGRP || mode & S_IXOTH) ) {
    fi->direct_io = 1;
  }
#endif

  if( ret == wfs_protocol::OK ) {
    fuse_reply_create( req, &e, fi );
  } else {
    fuse_reply_err( req, errno_from_status(ret) );
  }

  end_c_or_l();
  record_stat( statname, start );

}

void wfs_fuse_client_rename(fuse_req_t req, fuse_ino_t old_parent,
			    const char *old_name,
			    fuse_ino_t new_parent,
			    const char *new_name) {

  float start = start_stat();
  std::string statname = "rename";

  wfs_file_handle old_dh = get_fh( old_parent );
  wfs_file_handle new_dh = get_fh( new_parent );

  set_thread_id( old_dh, req );

  wfs_log(WFS_DBG_1, "Renaming from %s in directory %016"PRIx64" to %s in directory %016"PRIx64"\n",
	    old_name, old_dh.fid, new_name,
	    new_dh.fid );

  wfs_protocol::status ret = odc->rename( old_dh, old_name, new_dh, new_name,
					  req_ident(req));

  if( ret == wfs_protocol::OK ) {
    fuse_reply_err( req, 0 );
  } else {
    fuse_reply_err( req, errno_from_status(ret) );
  }
  clear_thread_id();
  record_stat( statname, start );

}

void wfs_fuse_client_mknod( fuse_req_t req, fuse_ino_t parent,
			    const char *name, mode_t mode, dev_t rdev ) {

  // TODO: Pay attention to mknods for devices?

  float start = start_stat();
  std::string statname = "mknod";

  struct fuse_entry_param e;
  start_c_or_l();
  int ret = create_helper( req, parent, name, mode, &e );

  if( !ret ) {
    fuse_reply_entry( req, &e );
  } else {
    fuse_reply_err( req, errno_from_status(ret) );
  }

  end_c_or_l();
  record_stat( statname, start );

}

void wfs_fuse_client_mkdir( fuse_req_t req, fuse_ino_t parent,
			    const char *name, mode_t mode ) {

  float start = start_stat();
  std::string statname = "mkdir";

  wfs_file_handle pfh = get_fh( parent );
  wfs_file_handle xfh;

  set_thread_id( pfh, req );

  struct stat st;
  wfs_log(WFS_DBG_1, "wfs_fuse_client_create %016"PRIx64" %s mode %o\n",
	    pfh.fid, name, mode );
  const struct fuse_ctx *ctx = fuse_req_ctx(req);
  wfs_protocol::status ret = odc->create( pfh, name,
					  permissions_for_mode(mode,
							       ctx->uid,
							       ctx->gid,
							       pfh,req,true),
					  WFS_TYPE_DIRECTORY,
					  req_ident(req), xfh );

  if( ret == wfs_protocol::OK &&
      (ret = getattr(req, xfh, st)) == wfs_protocol::OK ) {

    struct fuse_entry_param e;
    bzero( &e, sizeof(e) );
    e.ino = get_fuse_ino( xfh, parent );
    assert( e.ino );
    e.attr_timeout = 0.0;
    e.entry_timeout = 0.0;
    e.attr = st;
    fuse_reply_entry( req, &e );

  } else {
    fuse_reply_err( req, errno_from_status(ret) );
  }

  clear_thread_id();
  record_stat( statname, start );

}

void wfs_fuse_client_unlink( fuse_req_t req, fuse_ino_t parent,
			     const char *name ) {

  float start = start_stat();
  std::string statname = "unlink";

  wfs_file_handle pfh = get_fh( parent );
  set_thread_id( pfh, req );
  wfs_log(WFS_DBG_1, "wfs_fuse_client_unlink %016"PRIx64" %s\n",
	    pfh.fid, name );
  wfs_protocol::file_and_version fv;
  wfs_protocol::status ret = odc->unlink( pfh, name, false, true,
					  req_ident(req), fv );
  wfs_log(WFS_DBG_1, "unlink return value: %d\n", ret );

  if( ret == wfs_protocol::OK ) {
    fuse_reply_err( req, 0 );
  } else {
    fuse_reply_err( req, errno_from_status(ret) );
  }
  clear_thread_id();
  record_stat( statname, start );

}

void wfs_fuse_client_rmdir( fuse_req_t req, fuse_ino_t parent,
			    const char *name ) {

  float start = start_stat();
  std::string statname = "rmdir";

  wfs_file_handle pfh = get_fh( parent );
  set_thread_id( pfh, req );
  wfs_protocol::file_and_version fv;
  wfs_protocol::status ret = odc->unlink( pfh, name, true, true,
					  req_ident(req), fv );
  if( ret == wfs_protocol::OK ) {
    fuse_reply_err( req, 0 );
  } else {
    fuse_reply_err( req, errno_from_status(ret) );
  }
  clear_thread_id();
  record_stat( statname, start );

}

int getrss() {

  std::string filename = "getrss.";
  filename += random();

  char command[1000];
  sprintf( command, "cat /proc/%d/status | grep VmRSS | awk '{print $2}' > %s",
	   getpid(), filename.c_str() );
  system( command );

  int fd = open( filename.c_str(), O_RDONLY );
  char buf[100];
  read( fd, buf, 100 );
  close(fd);
  unlink( filename.c_str() );

  return atoi(buf);

}

void wfs_fuse_client_lookup( fuse_req_t req, fuse_ino_t parent,
			     const char *name ) {

  float start = start_stat();
  std::string statname = "lookup";

#ifdef GOOGLE_HEAP
  int rss = getrss();
  if( rss > 200000 ) {
    printf( "Resstd::string memory is %d KB\n",  rss );
    HeapProfilerDump( "none" );
  }
#endif

  struct fuse_entry_param e;
  wfs_file_handle fh = get_fh( parent );
  set_thread_id( fh, req );
  wfs_log(WFS_DBG_1, "got lookup for parent fid %016"PRIx64" and name %s\n",
	    fh.fid, name );
  wfs_file_handle xfh;
  wfs_protocol::status ret = wfs_protocol::NOTDIR;
  struct stat st;

  start_c_or_l();

  if( odc->isdir(fh) ) {

    ret = odc->lookup( fh, name, req_ident(req), xfh );
    if( ret == wfs_protocol::OK &&
	(ret = getattr(req, xfh, st, (fh.fid == xfh.fid))) == wfs_protocol::OK ) {

      wfs_log(WFS_DBG_1, "Lookup returned file handle: %s\n", xfh.print().c_str() );
      bzero( &e, sizeof(e) );
      e.ino = get_fuse_ino( xfh, parent );
      assert( e.ino );
      e.attr_timeout = 0.0;
      e.entry_timeout = 0.0;
      e.attr = st;
      wfs_log(WFS_DBG_1,"lookup for parent fid %016"PRIx64" and name %s finished: %016"PRIx64", %u\n",
          fh.fid, name, xfh.fid,
          (unsigned)e.ino );
      fuse_reply_entry(req, &e);
      end_c_or_l();
      clear_thread_id();
      record_stat( statname, start );
      return;

    }
  }
  fuse_reply_err( req, errno_from_status(ret) );
  end_c_or_l();
  clear_thread_id();
  record_stat( statname, start );
}

struct dirbuf {
    char *p;
    size_t size;
};


void dirbuf_add( fuse_req_t req, struct dirbuf *b, const char *name,
		 fuse_ino_t ino ) {

    struct stat stbuf;
    size_t oldsize = b->size;
    b->size += fuse_add_direntry(req, NULL, 0, name, NULL, 0);
    b->p = (char *) realloc(b->p, b->size);
    memset(&stbuf, 0, sizeof(stbuf));
    stbuf.st_ino = ino;
    fuse_add_direntry(req, b->p + oldsize, b->size - oldsize,
		      name, &stbuf, b->size);

}

#define min(x, y) ((x) < (y) ? (x) : (y))

int reply_buf_limited( fuse_req_t req, const char *buf, size_t bufsize,
		       off_t off, size_t maxsize ) {

  if( (size_t)off < bufsize ) {
    return fuse_reply_buf(req, buf + off, min(bufsize - off, maxsize));
  } else {
    return fuse_reply_buf(req, NULL, 0);
  }

}

void wfs_fuse_client_readdir( fuse_req_t req, fuse_ino_t ino, size_t size,
			      off_t off, struct fuse_file_info *fi ) {

  float start = start_stat();
  std::string statname = "readdir";

  wfs_file_handle fh = get_fh( ino );
  set_thread_id( fh, req );
  wfs_log(WFS_DBG_1, "going to readdir for %s %d %d\n", fh.print().c_str(),
	   (int)size, (int)off );

  struct dirbuf b;
  std::vector<wfs_storage::dirent> e;
  unsigned long long ind = 0;

  if( !odc->isdir(fh) ){
    fuse_reply_err( req, ENOTDIR );
    clear_thread_id();
    record_stat( statname, start );
    return;
  }

  memset( &b, 0, sizeof(b) );
  static const unsigned int MAX_DIR_READ = 1024;

  do {

    e.clear();
    wfs_protocol::status s = odc->readdir( fh, ind, MAX_DIR_READ,
					   req_ident(req), e );
    if( s != wfs_protocol::OK ) {
      // TODO: return error?  what about any previous data?
      break;
    }
    for( unsigned i = 0; i < e.size(); i++ ) {
      wfs_log(WFS_DBG_1, "adding entry %s to dir buf, inum %016"PRIx64"\n",
          e[i].name.c_str(), e[i].inum.fid );
      fuse_ino_t ino_new = get_fuse_ino(e[i].inum);
      assert( ino_new );
      dirbuf_add( req, &b, e[i].name.c_str(), ino_new );
    }
    ind += e.size();

  } while( e.size() == MAX_DIR_READ && b.size < (off+size) );

  reply_buf_limited( req, b.p, b.size, off, size );
  free(b.p);
  clear_thread_id();
  record_stat( statname, start );

}

void wfs_fuse_client_open( fuse_req_t req, fuse_ino_t ino,
			   struct fuse_file_info *fi ) {

  float start = start_stat();
  std::string statname = "open";

  wfs_file_handle fh = get_fh( ino );
  set_thread_id( fh, req );
  wfs_protocol::status ret;

  wfs_log(WFS_DBG_1, "Opening ino %d: %s\n", (int)ino, fh.print().c_str());

  if( !odc->isfile(fh) ) {
    fuse_reply_err( req, EISDIR );
    clear_thread_id();
    record_stat( statname, start );
    return;
  }

  // Linux doesn't keep track of the pointer for you when a file
  // is opened under O_APPEND; we have to track it ourselves somehow.
  // Set a flag in the fi->fh, since we're not using it for anything else.
  fi->fh = 0;
  if( fi->flags & O_APPEND ) {
    fi->fh |= O_APPEND;
  }

  // TODO: we may be doing more getattr calls than necessary (one in lookup
  // and one here) -- we should look into optimizing this later on.  It's
  // necessary here since we need to know the newest version number to read.

  struct stat st;
  if( (ret = getattr(req, fh, st )) == wfs_protocol::OK ) {

    // if the file is executable, we should not use direct_io
    // otherwise, we use direct_io to avoid the dreaded double read problem
    // on Linux, which messes with our maxtime timing.
#ifndef __APPLE__
    // !xavid: this should be changed to use permissions, because this won't
    //         always be accurate?
    if( odc->isfile(fh) &&
        !(st.st_mode & S_IXUSR || st.st_mode & S_IXGRP ||
	  st.st_mode & S_IXOTH) ) {
      // Hopefully this is ok on OSX? ~xavid: it's not, breaks append
      fi->direct_io = 1;
    }
#endif

    fuse_reply_open( req, fi );
  } else {
    fuse_reply_err( req, errno_from_status(ret) );
  }

  clear_thread_id();
  record_stat( statname, start );
}

void wfs_fuse_client_read( fuse_req_t req, fuse_ino_t ino, size_t size,
			   off_t off, struct fuse_file_info *fi ) {

  float start = start_stat();
  std::string statname = "read";

  wfs_file_handle fh = get_fh( ino );
  set_thread_id( fh, req );
  wfs_version version = get_ver( ino );
  wfs_protocol::status ret;
  std::string buf;

  do {

    ret = odc->readfile( fh, version, off, size, req_ident(req), buf );

    wfs_log(WFS_DBG_1, "Read called on file %016"PRIx64" ver %d, off=%"PRId64", size=%d ret=%d\n",
	      fh.fid, version.version_num, off, (int)size, ret);

    if( ret != wfs_protocol::OK ) {
      int err = errno_from_status(ret);
      if( ret == wfs_protocol::BADVERSION ) {
	// get the new version and try again.
	// TODO: limit iterations or time limit (or have actual versioning)
        struct stat st;
        wfs_log(WFS_DBG_1, "Tried to read bad version (%d) of %016"PRIx64"\n",
                  version.version_num, fh.fid );
        wfs_version oldver = version;
        fh.version.version_num = -1;

        if( (ret = getattr(req, fh, st )) != wfs_protocol::OK ) {
          wfs_log(WFS_DBG_1, "getattr after bad version failed: %d\n", ret );
          err = ESTALE;
        } else {
          version = get_ver( ino );
	  if (version == oldver) {
	    wfs_log(WFS_DBG_1, "A bad version was consistently requested, probably due to a .version= cue.\n");
	    err = ENOENT;
	  } else
	    continue;
        }
      } else {
        wfs_log(WFS_DBG_1,"read: Got a status %d!\n",(int)ret);
      }
      fuse_reply_err( req, err );
      clear_thread_id();
      record_stat( statname, start );
      return;
    }

    break;

  } while(1);

  fuse_reply_buf( req, buf.data(), buf.length() );
  clear_thread_id();
  record_stat( statname, start );
}

void wfs_fuse_client_flush( fuse_req_t req, fuse_ino_t ino,
			    struct fuse_file_info *fi ) {

  float start = start_stat();
  std::string statname = "flush";

  wfs_file_handle fh = get_fh( ino );
  set_thread_id( fh, req );

  wfs_log(WFS_DBG_1, "flush called on fh %016"PRIx64"\n", get_fh( ino ).fid );

  // we do the flushing here, rather than in close, since it's synchronous
  wfs_version new_version = get_ver( ino );

  // before we start the flush, pre-emptively increase the version number
  // this is to prevent a read for the same file coming in from another thread,
  // and looking at an old cached block that won't have the new writes applied
  // to it (since they're deleted before the flush RPC goes out)
  // Of course if that read request gets re-ordered with the flush,
  // the server won't yet know about this version yet (or worse, another
  // client has done a flush and gotten that version number to the server
  // first), but I'm in a hurry and don't know how else to solve this...
  wfs_version tmp_version = new_version;
  tmp_version.version_num++;
  set_ver( ino, tmp_version );

  wfs_protocol::status ret = odc->flush( fh, req_ident(req), new_version );

  if( ret == wfs_protocol::OK ) {
    // we do care about the new version number, so that our reads
    // will match our writes after an fsync
    set_ver( ino, new_version );
    fuse_reply_err( req, 0 );
  } else {
    wfs_log(WFS_DBG_1, "read.flush: Got a status %d!\n",(int)ret);
    fuse_reply_err( req, errno_from_status(ret) );
  }
  clear_thread_id();
  record_stat( statname, start );

}

void wfs_fuse_client_fsync( fuse_req_t req, fuse_ino_t ino, int datasync,
			    struct fuse_file_info *fi ) {

  float start = start_stat();
  std::string statname = "fsync";

  wfs_log(WFS_DBG_1, "fsync called, redirecting to flush\n" );

  // TODO: let's ignore the datasync parameter for now --
  // if it's non-zero, we're not supposed to flush metadata for some
  // reason.

  wfs_fuse_client_flush( req, ino, fi );
  record_stat( statname, start );

}

void wfs_fuse_client_setattr( fuse_req_t req, fuse_ino_t ino,
			      struct stat *attr, int to_set,
			      struct fuse_file_info *fi ) {

  float start = start_stat();
  std::string statname = "setattr";

  wfs_file_handle fh = get_fh( ino );
  set_thread_id( fh, req );

  if( wfs_dbg_block(WFS_DBG_1) ) {
    wfs_log(WFS_DBG_1, "setattr called on fh %016"PRIx64"", fh.fid );

    if( to_set & FUSE_SET_ATTR_SIZE ) {
      wfs_log(WFS_DBG_1, "\t size %"PRId64, attr->st_size);
    }
    if( to_set & FUSE_SET_ATTR_ATIME ) {
      wfs_log(WFS_DBG_1, "\t atime %.20g", (double)attr->st_atime);
    }
    if( to_set & FUSE_SET_ATTR_MTIME ) {
      wfs_log(WFS_DBG_1, "\t mtime %.20g", (double)attr->st_mtime);
    }
    if( to_set & FUSE_SET_ATTR_MODE ) {
      wfs_log(WFS_DBG_1, "\t mode %#o", attr->st_mode);
    }
    if( to_set & FUSE_SET_ATTR_UID ) {
      wfs_log(WFS_DBG_1, "\t uid %d", attr->st_uid);
    }
    if( to_set & FUSE_SET_ATTR_GID ) {
      wfs_log(WFS_DBG_1, "\t gid %d", attr->st_gid);
    }
  }

  wfs_protocol::status ret;

  if( to_set & FUSE_SET_ATTR_SIZE ) {
    wfs_version version = get_ver( ino );
    wfs_version new_version;
    ret = odc->truncate( fh, version, attr->st_size, req_ident(req),
			 new_version );
    if (ret != wfs_protocol::OK) {
      wfs_log(WFS_DBG_1, "setattr.truncate: Got a status %d!\n",(int)ret);
      fuse_reply_err( req, errno_from_status(ret) );
      clear_thread_id();
      record_stat( statname, start );
      return;
    }
  }

  if ( to_set & (FUSE_SET_ATTR_MODE
		 | FUSE_SET_ATTR_UID
		 | FUSE_SET_ATTR_GID)) {

    std::map<std::string, wfs_permission> perms;
    uid_t uid;
    gid_t gid;
    mode_t mode;

    if ( to_set & FUSE_SET_ATTR_UID ) {
      uid = attr->st_uid;
    } else
      uid = -1;
    if ( to_set & FUSE_SET_ATTR_GID ) {
      gid = attr->st_gid;
    } else
      gid = -1;

    if( to_set & FUSE_SET_ATTR_MODE ) {
      mode = attr->st_mode;
    } else
      mode = -1;

    perms = permissions_for_mode(mode,
				 uid,
				 gid,
				 fh,
				 req,
				 false);

    wfs_version version = get_ver( ino );
    wfs_version new_version;
    // xavid: chmod does the right thing and doesn't modify permissions
    //        we don't specify here
    ret = odc->chmod( fh, perms, req_ident(req), new_version );

    if (ret != wfs_protocol::OK) {
      wfs_log(WFS_DBG_1, "setattr.chmod: Got a status %d!\n",(int)ret);
      fuse_reply_err( req, errno_from_status(ret) );
      clear_thread_id();
      record_stat( statname, start );
      return;
    }
  }

  struct stat buf;

  ret = getattr(req, fh, buf);
  if (ret != wfs_protocol::OK) {
    wfs_log(WFS_DBG_1, "setattr.getattr2: Got a status %d!\n",(int)ret);
    fuse_reply_err( req, errno_from_status(ret) );
    clear_thread_id();
    record_stat( statname, start );
    return;
  }

  fuse_reply_attr( req, &buf, 0.0 );
  clear_thread_id();
  record_stat( statname, start );
}

void wfs_fuse_client_write( fuse_req_t req, fuse_ino_t ino,
			    const char *buf, size_t size, off_t off,
			    struct fuse_file_info *fi ) {

  float start = start_stat();
  std::string statname = "write";

  wfs_file_handle fh = get_fh( ino );
  set_thread_id( fh, req );

  wfs_log(WFS_DBG_1, "wfs_fuse_client_write %d bytes\n", (int) size);

  wfs_version version = get_ver( ino );
  wfs_protocol::status ret;

  wfs_version new_version;

  if( fi->fh & O_APPEND ) {
    struct stat buf;
    // As long as we're using a client that has caching, this shouldn't
    // add any latency
    getattr(req, fh, buf );
    off = buf.st_size;
  }

  const std::string data = std::string(buf, size);

  ret = odc->writefile( fh, version, off, data, req_ident(req), new_version );
  wfs_log(WFS_DBG_1, "odc->writefile(%016"PRIx64" off %d size %d) ret %d ver %d->%d\n",
	    fh.fid, (int)off, (int)size,
            ret, version.version_num, new_version.version_num);
  set_ver( ino, new_version );

  if( ret == wfs_protocol::OK ) {
    fuse_reply_write( req, size );
  } else {
    wfs_log(WFS_DBG_1, "write: Got a status %d!\n",(int)ret);
    fuse_reply_err( req, errno_from_status(ret) );
  }
  clear_thread_id();
  record_stat( statname, start );

}

void wfs_fuse_client_release( fuse_req_t req, fuse_ino_t ino,
			      struct fuse_file_info *fi ) {

  float start = start_stat();
  std::string statname = "release";

  if( ino == 0 ) {
    wfs_log(WFS_DBG_1, "Ignoring release for inode of 0\n" );
    fuse_reply_err( req, 0 );
    record_stat( statname, start );
    return;
  }

  wfs_log(WFS_DBG_1, "release called on fh %016"PRIx64"\n", get_fh( ino ).fid );

#if defined(__linux__) || defined(__APPLE__)
  // flush does all the work
  fuse_reply_err( req, 0 );
#else
  // FreeBSD doesn't have flush, but its release seems to be
  // synchronous?
  if( fi->flush ) {
    wfs_fuse_client_flush(req, ino, fi);
  }
#endif

  record_stat( statname, start );

}

void wfs_fuse_client_readlink( fuse_req_t req, fuse_ino_t ino ) {

  float start = start_stat();
  std::string statname = "readlink";

  wfs_file_handle fh = get_fh( ino );
  set_thread_id( fh, req );
  wfs_version version = get_ver( ino );

  wfs_log(WFS_DBG_1, "wfs_fuse_client_readlink %016"PRIx64"\n", fh.fid );

  std::string link;
  wfs_protocol::status ret = odc->readlink( fh, version, req_ident(req),link );

  if( ret == wfs_protocol::OK ) {
    fuse_reply_readlink( req, link.c_str() );
  } else {
    wfs_log(WFS_DBG_1, "readlink: Got a status %d!\n",(int)ret);
    fuse_reply_err( req, errno_from_status(ret) );
  }
  clear_thread_id();
  record_stat( statname, start );

}

void wfs_fuse_client_symlink( fuse_req_t req, const char *link,
			      fuse_ino_t parent, const char *name ) {

  float start = start_stat();
  std::string statname = "symlink";

  wfs_file_handle fh = get_fh( parent );
  set_thread_id( fh, req );

  wfs_log(WFS_DBG_1, "wfs_fuse_client_symlink in %016"PRIx64" bytes, %s -> %s\n",
	    fh.fid, name, link );

  wfs_file_handle new_fh;
  wfs_protocol::status ret = odc->symlink( fh, std::string(name),
					   std::string(link),
					   req_ident(req), new_fh );

  struct stat st;
  if( ret == wfs_protocol::OK &&
      (ret = getattr(req, new_fh, st)) == wfs_protocol::OK ) {

    struct fuse_entry_param e;
    bzero( &e, sizeof(e) );
    e.ino = get_fuse_ino( new_fh, parent );
    assert( e.ino );
    e.attr_timeout = 0.0;
    e.entry_timeout = 0.0;
    e.attr = st;
    fuse_reply_entry( req, &e );

  } else {
    wfs_log(WFS_DBG_1, "symlink: Got a status %d!\n",(int)ret);
    fuse_reply_err( req, errno_from_status(ret) );
  }
  clear_thread_id();
  record_stat( statname, start );

}

void wfs_fuse_client_link( fuse_req_t req, fuse_ino_t ino,
			   fuse_ino_t newparent, const char *newname ) {

  float start = start_stat();
  std::string statname = "link";

  wfs_file_handle fh = get_fh( ino );
  wfs_file_handle parent_fh = get_fh( newparent );
  set_thread_id( fh, req );

  wfs_log(WFS_DBG_1, "wfs_fuse_client_link to %016"PRIx64" in dir %016"PRIx64", new name %s\n",
	    fh.fid, parent_fh.fid, newname );

  wfs_file_handle new_fh;
  wfs_version new_version;
  // gives 0 permissions for some reason?
  wfs_protocol::status ret = odc->link( fh, parent_fh, std::string(newname),
					false, true, WFS_TYPE_REGULAR,
					std::map<std::string,
					wfs_permission>(),
					req_ident(req), new_fh, new_version );

  struct stat st;
  if( ret == wfs_protocol::OK &&
      (ret = getattr(req, fh, st)) == wfs_protocol::OK ) {

    struct fuse_entry_param e;
    bzero( &e, sizeof(e) );
    e.ino = get_fuse_ino( fh, newparent );
    assert( e.ino );
    e.attr_timeout = 0.0;
    e.entry_timeout = 0.0;
    e.attr = st;
    fuse_reply_entry( req, &e );

  } else {
    wfs_log(WFS_DBG_1, "link: Got a status %d!\n",(int)ret);
    fuse_reply_err( req, errno_from_status(ret) );
  }
  clear_thread_id();
  record_stat( statname, start );

}

void wfs_fuse_client_forget( fuse_req_t req, fuse_ino_t ino,
			     unsigned long nlookup ) {

  wfs_log(WFS_DBG_1, "forget called\n" );
  fuse_reply_none(req);

}

void wfs_fuse_client_statfs( fuse_req_t req, fuse_ino_t ) {

    struct statvfs buf;

    memset(&buf, 0, sizeof(buf));

    buf.f_namemax = 255;
    buf.f_bsize = 512;

    fuse_reply_statfs(req, &buf);

}

struct fuse_lowlevel_ops wfs_fuse_client_oper;

int init_and_run_wfs_fuse_client( wfs_options *opts, vivaldi *viv,
				  slice_manager *slices,
				  std::string progname,
				  wfs_client_manager *clients,
				  wfs_server *server = NULL ) {

#ifdef GOOGLE_HEAP
  HeapProfilerStart( "/tmp/wfs-heap" );
#endif

  int err = -1;
  struct sockaddr_in dst;

  bzero( &dst, sizeof(dst) );

  assert( pthread_mutex_init( &fh_to_ino_lock, NULL ) == 0 );
  assert( pthread_mutex_init( &stats_lock, NULL ) == 0 );
  assert( pthread_mutex_init( &create_lock, NULL ) == 0 );
  assert( pthread_cond_init( &create_condvar, NULL ) == 0 );

  wfs_log(WFS_DBG_1, "Local node is %s:%d\n", opts->ip.c_str(), opts->port );

  wfs_cache *cache = new wfs_cache_mem( opts->cache_dir,
					100*1024*1024 /*100M*/ );
  //wfs_cache *cache = new wfs_cache_blockfile( cache_dir );

  sockaddr_in client_addr;
  inet_aton( opts->ip.c_str(), &client_addr.sin_addr );
  client_addr.sin_port = htons( opts->client_port );

  rpcs *client_server = new rpcs( client_addr.sin_port );
  if( client_addr.sin_port == 0 ) {
    client_addr.sin_port = htons(client_server->bound_port());
    wfs_log(WFS_DBG_1, "Client bound to port %d\n", client_server->bound_port() );
  }

  wfs_node me = opts->me;
  if( me.id == 0 ) {

    // find a home node
    std::vector<host> hosts;
    slices->get_all_active_hosts(hosts);
    unsigned int rand = random() % hosts.size();
    struct in_addr addr;
    inet_aton(opts->ip.c_str(), &addr);
    unsigned int local_ip = addr.s_addr;
    unsigned int curr_ip = 0;
    std::vector<host> same_ip;
    std::vector<host> same_site;

    for( unsigned int i = 0; i < hosts.size(); i++ ) {
      // pick the node according to these preferences:
      //   1) Has the same IP address
      //   2) Has the same site
      //   3) Is a random host
      if( me.id == 0 && i == rand ) {
	me = make_node(hosts[i]);
	curr_ip = hosts[i].addr;
      } else if( local_ip == hosts[i].addr ) {
	same_ip.push_back(hosts[i]);
      } else if( hosts[i].site_id == opts->site_id ) {
	same_site.push_back(hosts[i]);
      }
    }

    if( curr_ip != local_ip && same_ip.size() > 0 ) {
      rand = random() % same_ip.size();
      me = make_node(same_ip[rand]);
    } else if( curr_ip != local_ip && same_site.size() > 0 ) {
      rand = random() % same_site.size();
      me = make_node(same_site[rand]);
    }

    wfs_log(WFS_DBG_1, "Using %s:%d as my local node\n", inet_ntoa(me.addr.sin_addr),
	      ntohs(me.addr.sin_port) );

  }

  odc = new wfs_bt_client( client_addr, me, slices, cache,
			   client_server, viv, clients, opts->read_ahead,
			   opts->prefetch_threads );
  odc->set_local_server( server );
  wfs_log(WFS_DBG_1, "Starting client's BT at port %s:%d\n",
	    opts->ip.c_str(), client_server->bound_port() );

  if( wfs_dbg_block(WFS_DBG_1) ) {
    int32_t mask = random() & 0xffff0000;
    wfs_thread_id_init(mask);
    wfs_log(WFS_DBG_1, "Mask is %08x\n", mask );
  }

  no_permissions = opts->no_permissions;
  if( no_permissions ) {
    wfs_log(WFS_DBG_1, "Disabling permissions checking\n" );
  }

  wfs_fuse_client_oper.lookup  = wfs_fuse_client_lookup;
  wfs_fuse_client_oper.getattr = wfs_fuse_client_getattr;
  wfs_fuse_client_oper.setattr = wfs_fuse_client_setattr;
  wfs_fuse_client_oper.readdir = wfs_fuse_client_readdir;
  wfs_fuse_client_oper.open    = wfs_fuse_client_open;
  wfs_fuse_client_oper.read    = wfs_fuse_client_read;
  wfs_fuse_client_oper.write   = wfs_fuse_client_write;
  wfs_fuse_client_oper.statfs  = wfs_fuse_client_statfs;
  wfs_fuse_client_oper.create  = wfs_fuse_client_create;
  wfs_fuse_client_oper.mknod   = wfs_fuse_client_mknod;
  wfs_fuse_client_oper.mkdir   = wfs_fuse_client_mkdir;
  wfs_fuse_client_oper.fsync   = wfs_fuse_client_fsync;
  wfs_fuse_client_oper.unlink  = wfs_fuse_client_unlink;
  wfs_fuse_client_oper.rmdir   = wfs_fuse_client_rmdir;
  wfs_fuse_client_oper.release = wfs_fuse_client_release;
  wfs_fuse_client_oper.flush   = wfs_fuse_client_flush;
  wfs_fuse_client_oper.rename  = wfs_fuse_client_rename;
  wfs_fuse_client_oper.symlink = wfs_fuse_client_symlink;
  wfs_fuse_client_oper.readlink= wfs_fuse_client_readlink;
  wfs_fuse_client_oper.link    = wfs_fuse_client_link;

  char *fuse_argv[20];
  int fuse_argc = 0;
  fuse_argv[fuse_argc++] = (char *) progname.c_str();
#ifdef __APPLE__
  //fuse_argv[fuse_argc++] = "-o";
  //fuse_argv[fuse_argc++] = "novncache"; // no dir entry caching
  // ~xavid: disabling all local caches helps pass tests
  fuse_argv[fuse_argc++] = "-o";
  fuse_argv[fuse_argc++] = "nolocalcaches";
  fuse_argv[fuse_argc++] = "-o";
  fuse_argv[fuse_argc++] = "daemon_timeout=86400";
#endif

  // On FreeBSD 6 (using FUSE 2.6.3), I can only mount as root.
  // This option allows me to mount FUSE as root, but access
  // the mounted directory as another user.
  // (Don't know if you need this on Mac or not, but on Linux you
  // can just mount as yourself.)

  // everyone can play, why not?
  fuse_argv[fuse_argc++] = (char *)"-o";
  fuse_argv[fuse_argc++] = (char *)"allow_other";

  fuse_argv[fuse_argc++] = opts->mountpoint;
  if( wfs_dbg_block(2) ) {
    fuse_argv[fuse_argc++] = (char *)"-d";
  }

  fuse_args args = FUSE_ARGS_INIT( fuse_argc, fuse_argv );
  int foreground;
  int res = fuse_parse_cmdline( &args, &opts->mountpoint, 0 /*multithreaded*/,
				&foreground );
  wfs_log(WFS_DBG_1, "Mounting at %s\n", opts->mountpoint );

  if( res == -1 ) {
    return 0;
  }

  args.allocated = 0;

  struct fuse_chan *ch;
  ch = fuse_mount( opts->mountpoint, &args );
  if( ch == NULL ) {
    wfs_log(WFS_DBG_OFF, "fuse_mount failed %s\n", opts->mountpoint );
    exit(1);
  }

  struct fuse_session *se;

  se = fuse_lowlevel_new( &args, &wfs_fuse_client_oper,
			  sizeof(wfs_fuse_client_oper),
			  NULL );
  if( se == 0 ) {

    wfs_log(WFS_DBG_OFF, "fuse_lowlevel_new failed\n" );
    exit(1);
  }

#ifdef WFS_STATS
  signal( SIGINT, &print_stats );
  signal( SIGKILL, &print_stats );
#endif

  fuse_session_add_chan( se, ch );
  err = fuse_session_loop_mt(se);
  //err = fuse_session_loop(se);

  fuse_unmount(opts->mountpoint, ch);
  fuse_session_destroy(se);

  return err ? 1 : 0;

}

