/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <getopt.h>
#include <rpc.h>
#include <wfs_file.h>
#include <wfs_misc.h>
#include <sys/time.h>

#include "wfs_options.h"
#include "util/ident_util.h"

wfs_options::wfs_options( bool server, bool client ) :
  _server(server), _client(client) {

  addr = NULL;
  port = 0;
  site_id = "";
  port_char = NULL;
  cserver_master = NULL;
  dir = NULL;
  debug_level = 0;
  secure = false;
  username = server?"host":current_username();
  key_dir = default_keydir;
  ping_period = 10;
  //ping_period = 60;  // for PlanetLab
  maint_timer = 10;
  cache_dir = "/tmp/wfs_cache";
  client_port = 0;
  mountpoint = NULL;
  no_permissions = false;
  read_ahead = true;
  prefetch_threads = 5;
  dbg_dir = ".";
}

void wfs_options::usage( std::string progname ) {

  const std::string client =
    "\t--mountpoint <path>\n"
    "\t--cache-dir <path>\n"
    "\t--client-port <port>\n"
#ifdef WFS_WITH_LIBSSH
    "\t[--ident <username>]\n"
#endif
    "\t--read-ahead\n"
    "\t--prefetch-threads <num>\n";

  const std::string server =
    "\t--dir <dir>\n";

  const std::string both =
    "\t--addr <ip>\n"
    "\t--port <port>\n"
    "\t--site <site_id>\n"
    "\t--cserv-master <ip:port>\n"
    "\t--ping-period <seconds>\n"
    "\t--maint-timer <seconds>\n"
    "\t--no-permissions\n"
#ifdef WFS_WITH_LIBSSH
    "\t[--secure]\n"
    "\t[--keydir <path>] (defaults to ~/.wfs/)\n"
#endif
    "\t--debug (may be given multiple times)\n"
    "\t--dbg-dir <debug_log_directory>\n";

  std::string usage = both;
  if( _server ) {
    usage += server;
  }
  if( _client ) {
    usage += client;
  }

  fprintf( stderr, "Usage for %s:\n\n%s",
	   progname.c_str(),
	   usage.c_str()
	   );

  exit(-1);

}

static void add_opt( struct option *opt, const char *name, int arg, char id,
		     std::string &os ) {
  opt->name = name;
  opt->has_arg = arg;
  opt->flag = 0;
  opt->val = id;
  if( name != 0 ) {
    os += id;
    if( arg == required_argument ) {
      os += ':';
    }
  }
}

void wfs_options::get_options( int argc, char *argv[] ) {

  int c;
  optind = 0;  // reset getopt

  struct option long_options[30];
  int i = 0;
  std::string os = "";
  add_opt( &(long_options[i++]), "cserv-master", required_argument, 'c', os );
  add_opt( &(long_options[i++]), "ping-period", required_argument, 't', os );
  add_opt( &(long_options[i++]), "maint-timer", required_argument, 'M', os );
  add_opt( &(long_options[i++]), "debug", no_argument, 'd', os );
  add_opt( &(long_options[i++]), "dbg-dir", required_argument, 'b', os );
  add_opt( &(long_options[i++]), "no-permissions", no_argument, 'n', os );
  add_opt( &(long_options[i++]), "addr", required_argument, 's', os );
  add_opt( &(long_options[i++]), "port", required_argument, 'p', os );
  add_opt( &(long_options[i++]), "site", required_argument, 'e', os );
#ifdef WFS_WITH_LIBSSH
  add_opt( &(long_options[i++]), "secure", no_argument, 'S', os );
  add_opt( &(long_options[i++]), "keydir", required_argument, 'K', os );
#endif
  add_opt( &(long_options[i++]), "help", no_argument, 'h', os );

  if( _server ) {
    add_opt( &(long_options[i++]), "dir", required_argument, 'i', os );
  }

  if( _client ) {
    add_opt( &(long_options[i++]), "mountpoint", required_argument, 'm', os );
    add_opt( &(long_options[i++]), "cache-dir", required_argument, 'a', os );
    add_opt( &(long_options[i++]), "client-port", required_argument, 'l', os);
    add_opt( &(long_options[i++]), "read-ahead", no_argument, 'r', os);
    add_opt( &(long_options[i++]), "prefetch-threads",required_argument, 'f', os);
#ifdef WFS_WITH_LIBSSH
    add_opt( &(long_options[i++]), "ident", required_argument, 'I', os );
#endif
  }

  add_opt( &(long_options[i++]), 0, 0, 0, os );

  while(1) {

    int option_index = 0;
    c = getopt_long( argc, argv, os.c_str(), long_options, &option_index );

    if( c == -1 ) {
      break;
    }

    switch(c) {

    case 's':
      addr = optarg;
      break;

    case 'p':
      port = atoi(optarg);
      port_char = optarg;
      break;

    case 'e':
      site_id = std::string(optarg);
      if( site_id == WFS_RANDOM_SITE ) {
        fprintf( stderr, "%s is a reserved site name; you can't use it!\n",
            site_id.c_str() );
	exit(-1);
      }
      break;

    case 'i':
      dir = optarg;
      break;

    case 'n':
      no_permissions = true;
      break;

    case 'c':
      cserver_master = optarg;
      break;

    case 't':
      ping_period = atoi(optarg);
      break;

    case 'M':
      maint_timer = atoi(optarg);
      break;

    case 'a':
      cache_dir = optarg;
      break;

    case 'l':
      client_port_char = optarg;
      client_port = atoi(optarg);
      break;

    case 'm':
      mountpoint = optarg;
      break;

    case 'b':
      dbg_dir = optarg;
      break;

    case 'd':
      debug_level++;
      break;

#ifdef WFS_WITH_LIBSSH
    case 'S':
      secure = true;
      break;
    case 'K':
      key_dir = optarg;
      break;
    case 'I':
      if (!_server) {
	username = optarg;
      } else {
        fprintf(stderr, "You can't manually set an ident on a server!");
        exit(-1);
      }
      break;
#endif
    case '?':
    case 'h':
      /* getopt_long already printed an error message. */
      usage( std::string(argv[0]) );
      break;
    case 'r':
      read_ahead = true;
      break;
    case 'f':
      if (atoi(optarg) > 0) {
	prefetch_threads = atoi(optarg);
      }
      break;
    default:
      printf( "Unknown flag %c\n", c );
      usage( std::string(argv[0]) );
    }
  }

  if( (_client && mountpoint == NULL) ||
      (_server && dir == NULL) ||
      (_server && port == 0) ||
      cserver_master == NULL ) {
    fprintf( stderr, "Required parameter is missing.\n" );
    usage( std::string(argv[0]) );
  }

}

void wfs_options::standard_init() {

  fflush(stdout);
  setlinebuf(stdout);

  ip = std::string(get_local_ip());
  if( _server ) {
    if( !addr ) {
      me_host = host( ip.c_str(), port_char, site_id );
    } else {
      ip = std::string(addr);
      me_host = host( addr, port_char, site_id );
    }
    me = make_node(me_host);
  } else {
    me_host = host( ip.c_str(), client_port_char, site_id );
  }

  std::string name;
  if( _server && _client ) {
    name = "wfs";
  } else if( _server ) {
    name = "wfs_server";
  } else {
    name = "wfs_fuse_client";
  }

  TidWFSLogAugmentor * taug = new TidWFSLogAugmentor();
  _loginit = new WFSLogInit( name + "." + me_host.str(), dbg_dir, 
			     100, 5, taug );

  if( debug_level > 0 ) {
    wfs_set_debug( debug_level );
    wfs_log( WFS_DBG_1, "DEBUG LEVEL: %d\n", debug_level );
  }

  if (secure) {
    wfs_log(WFS_DBG_1, "enabling security as %s; using %s as keydir\n",
	    username.c_str(), key_dir);
    rpc_enable_security(wfs_public_key_identifies_user,
			wfs_get_private_key,
			username.c_str());
    wfs_set_key_dir(key_dir);
  }

  if( _server ) {
    wfs_log(WFS_DBG_1, "My node is %s:%d,%016"PRIx64"\n", ip.c_str(), 
	    port, me.id );
  }

  struct timeval tv;
  gettimeofday(&tv, NULL);
  srandom( tv.tv_usec^me.addr.sin_addr.s_addr^client_port );

}
