/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef chan_h
#define chan_h 1

//#define WFS_RPC_CHECKSUMMING 1

// channel abstraction that is used by rpcc and rpcs to communicate.
// this layer understands about connections and multiplexing many send()s
// simultaneously over a connection.
// it uses TCP connections to get good wide-area performance, etc.

#include <string>
#include <list>
#include <map>
#include <vector>
#include <netinet/tcp.h>

#include "base_chan.h"
#include "fifo.h"

// for rpcs.  one rpcs may have may have many tcpchans, one per rpcc
// that is communitcating with this rpcs.
class schan {
 private:
  struct inbuf;
  int channo;
  bool waiting;
  pthread_cond_t outstanding_c;
  chan_hub *hub;
  canceller cancel;

  int lossy_percent;

 public:
  schan(chan_hub *_hub, int port);
  ~schan();

  class cleanup {
  public:
    cleanup() {};
    virtual ~cleanup() {};
    virtual void do_cleanup(int channo) = 0;
  };

  void recv(charbuf &cb, int &channo);
  void send(const charbuf pdu, int channo);
  void done(pthread_t myth);
  bool set_cleanup(int channo, cleanup* obj);

  void lock_for_delete(int channo);
  void unlock_for_delete(int channo);

  void setlossy(int p=5) { lossy_percent = p; }

  bool verify_user(const char *username, int channo);

  unsigned int bound_port() { return tcp->bound_port(); }

 private:
  abstract_bind *tcp;

  // map channo to tcpchan (and its associated polling thread)
  struct channel {
    pthread_t th;
    abstract_channel *ch;
    cleanup *clnup;
    int outstanding;
    bool die;
  };
  std::map<int, channel> tcpchans;
  std::map<int, pthread_mutex_t *> delete_locks;
  std::list<int> lru_sockets;
  pthread_mutex_t tcpchans_m;
  pthread_cond_t tcpchans_c;

  // input queue. fed by a loop1() per tcpchan.
  // read & waited for by rpc's calls to recv().
  struct inbuf {
    inbuf(charbuf xs, int xchan) : s(xs), channo(xchan) { }
    inbuf() : channo(0) { }
    charbuf s;
    int channo;
  };
  fifo<inbuf> inq;
  
  pthread_t th_tcp_loop;
  void tcp_loop();
  void tcp_loop1(abstract_channel *s, pthread_mutex_t *delete_lock);
  static void cleanup_tcp_loop1(void *arg) {
    schan *sch = (schan *) arg;
    sch->done(pthread_self());
  }
  static void cleanup_lock(void *arg) {
    pthread_mutex_t *l = (pthread_mutex_t *) arg;
    pthread_mutex_unlock(l);
  };
  void send_tcp(int key, const charbuf &pdu);
  int find_him(sockaddr_in him);
};

// for rpcc
class cchan {

 private:
  pthread_mutex_t connect_m;
  pthread_mutex_t ch_m;
  pthread_cond_t ch_c;
  pthread_mutex_t tcpdelete_m;
  pthread_mutex_t setup_m;
  std::vector<pthread_t> setup_ths;
  canceller cancel;
  bool *deleted;

  pthread_t th;
  sockaddr_in dst;
  abstract_channel *ch;
  chan_hub *hub;
  fifo<charbuf> inq;
  void setup_and_send(const charbuf pdu);
  void tcp_loop();
  static void cleanup_tcp_loop(void *arg) {
    cchan *cch = (cchan *) arg;
    cch->done();
  }
  static void cleanup_lock(void *arg) {
    pthread_mutex_t *l = (pthread_mutex_t *) arg;
    pthread_mutex_unlock(l);
  }
  struct cleanup_lock_del {
    pthread_mutex_t *lock;
    bool *deleted;
  };
  static void cleanup_lock_delcheck(void *arg) {
    cleanup_lock_del *cld = (cleanup_lock_del *) arg;
    if(!*(cld->deleted)) {
      pthread_mutex_unlock(cld->lock);
    }
    delete cld;
  }
  struct cleanup_data {
    pthread_mutex_t *setup_m;
    std::vector<pthread_t> *setup_ths;
    pthread_t pth;
    bool *deleted;
  };
  static void cleanup_setup_thread(void *arg) {
    cleanup_data *cd = (cleanup_data *) arg;
    cchan::cleanup_setup_thread(cd->setup_m, cd->setup_ths, cd->pth, 
				cd->deleted);
    delete cd;
  }
  static void cleanup_close_channel(void *arg) {
    abstract_channel *c = (abstract_channel *) arg;
    // Turns out closing in the destructor does the wrong thing, because
    // the destructor tries to cancel the same thread that was canceled
    // to cause this function to be called.
    c->close();
    delete c; 
  }

  int lossy_percent;

 public:
  cchan(chan_hub *_hub, sockaddr_in _dst);
  ~cchan();
  void send(const charbuf pdu);
  void recv(charbuf &cb);
  bool dead();
  void done();
  static void cleanup_setup_thread(pthread_mutex_t *setup_m, 
				   std::vector<pthread_t> *setup_ths, 
				   pthread_t pth, bool *deleted);
  /**
   * At some point I'd love to have a lock per message or something,
   * but that seems way to expensive in terms of memory overhead
   * at the moment.
   */
  void lock_for_delete() { pthread_mutex_lock( &tcpdelete_m ); };
  void unlock_for_delete() { pthread_mutex_unlock( &tcpdelete_m ); };

  void setlossy(int p=5) { lossy_percent = p; }

  int initial_timeout() { return hub->initial_timeout(); }
};

#endif
