/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "sshchan.h"

#include "rpc.h"

#include "method_thread.h"
#include "logging/wfs_log.h"

#ifdef WFS_WITH_LIBSSH
#include "libssh/libssh.h"
#include "libssh/priv.h"

const int sshloglevel = SSH_LOG_PROTOCOL;

sshchan::sshchan(pthread_mutex_t *sshdelete_m,
		 bool freebuf, canceller *cancel,
		 SSH_SESSION *_session, int _channo)
  : abstract_channel(sshdelete_m, freebuf, cancel), chan(NULL),
    session(_session), fd(_channo)
{
  if (wfs_dbg_block(3)) {
    ssh_set_verbosity(sshloglevel);
  }
}

sshchan::~sshchan()
{
  wfs_log(WFS_DBG_2, "~sshchan socket %d\n", channo() );
  assert(pthread_cancel(th) == 0);
  assert(pthread_join(th, NULL) == 0);

  close();
  wfs_log(WFS_DBG_2, "~sshchan done socket %d\n", channo() );
}

int sshchan::start() {

  SSH_OPTIONS *opt = ssh_options_new();
  if (wfs_dbg_block(3)) {
    ssh_options_set_log_verbosity(opt, sshloglevel);
  }
  if(session->options) {
    ssh_options_free(session->options);
  }
  ssh_set_options(session,opt);

  if (ssh_accept(session)) {
    wfs_log(WFS_DBG_OFF,
	    "ssh_accept error: %s\n",
	    ssh_get_error(session));
    return -1;
  }
  if (!session->alive) {
    wfs_log(WFS_DBG_OFF,
	    "ssh_accept accepted a dead session with error: %s\n",
	    ssh_get_error(session));
    return -1;
  }

  SSH_MESSAGE *message;

  // First, we do auth
  // TODO: should we define these constants somewhere that breaks the
  // abstraction barrier less?
  char authuser[MAX_IDENT_LENGTH+1] = "";
  bool sent_banner = false;
  do {
    wfs_log(WFS_DBG_2, "sshchan auth waiting for a message\n");
    message = ssh_message_get(session);
    wfs_log(WFS_DBG_2, "sshchan auth got something!\n");
    if (!message) {
      wfs_log(WFS_DBG_OFF,
	      "sshchan auth message error: %s\n",ssh_get_error(session));
	return -1;
    }
    wfs_log(WFS_DBG_2, "sshchan auth got a message type %d subtype %d\n",
		       ssh_message_type(message),
		       ssh_message_subtype(message));

    if (!sent_banner) {
      // This is a hack; this banner should actually be an option
      // that you set ahead of time and let libssh send at the apropiate
      // time.
      // Pass our ident as the issue banner
      ssh_send_issue_banner(session,rpc_primary_user());
      sent_banner = true;
    }

    switch (ssh_message_type(message)) {
    case SSH_AUTH_REQUEST:
      switch (ssh_message_subtype(message)) {
      case SSH_AUTH_PUBLICKEY:
	switch (ssh_message_auth_phase(message)) {
	case PHASE_PUBLICKEY_START:
	  if (rpc_public_key_identifies_user(ssh_message_auth_user(message),
					ssh_message_auth_publickey(message))) {
	    wfs_log(WFS_DBG_2,
			       "sshchan auth requesting a sig for user %s\n",
			       ssh_message_auth_user(message));
	    ssh_message_auth_reply_request_signature(message);
	  } else {
	    // Default means failure
	    wfs_log(WFS_DBG_2, "sshchan auth failed start phase\n");
	    ssh_message_reply_default(message);
	  }
	  break;
	case PHASE_PUBLICKEY_SIGNATURE:
	  if (rpc_public_key_identifies_user(ssh_message_auth_user(message),
					 ssh_message_auth_publickey(message))
	      && ssh_message_auth_validated(message)) {
	    strncpy(authuser, ssh_message_auth_user(message),
		    sizeof(authuser)-1);
	    authuser[sizeof(authuser)-1] = '\0';
	    wfs_log(WFS_DBG_2,
			       "sshchan auth authenticated as user %s\n",
			       authuser);
	    ssh_message_auth_reply_success(message,0);
	  } else {
	    // Default means failure
	    wfs_log(WFS_DBG_2, "sshchan auth failed signature phase\n");
	    ssh_message_reply_default(message);
	  }
	  break;
	default:
	  wfs_log(WFS_DBG_OFF, "sshchan auth got an unknown phase %d\n",
		  ssh_message_auth_phase(message));
	  ssh_message_reply_default(message);
	  break;
	}
	break;
	//fprintf( stderr, "tried to use sshchan with stock libssh\n" );
	//assert(false);
      default:
        wfs_log(WFS_DBG_2, "sshchan auth tried to use something other than publickey\n");
	ssh_message_auth_set_methods(message,SSH_AUTH_PUBLICKEY);
	ssh_message_reply_default(message);
	break;
      }
      break;
    default:
      wfs_log(WFS_DBG_2, "stderr auth tried to do other stuff before authing\n");
      ssh_message_reply_default(message);
    }
    ssh_message_free(message);
  } while (authuser[0] == '\0');

  strncpy(authorized_user, authuser,
	  sizeof(authorized_user)-1);
  authorized_user[sizeof(authorized_user)-1] = '\0';

  chan = NULL;
  do {
    message = ssh_message_get(session);
    if (!message) {
      wfs_log(WFS_DBG_OFF,
	      "chan message error: %s\n",ssh_get_error(session));
      return -1;
    }
    switch (ssh_message_type(message)) {
      case SSH_CHANNEL_REQUEST_OPEN:
	if (ssh_message_subtype(message) == SSH_CHANNEL_SESSION) {
	  chan = ssh_message_channel_request_open_reply_accept(message);
	  wfs_log(WFS_DBG_2, "sshchan sopen got an open\n");
	  break;
	} else {
	  wfs_log(WFS_DBG_2, "sshchan sopen got an unhandled open type: %d\n", ssh_message_subtype(message));
	  /* don't break */
	}
    default:
      wfs_log(WFS_DBG_2, "sshchan sopen got an unhandled message expecting an open: %d\n", ssh_message_type(message));
      ssh_message_reply_default(message);
      break;
    }
    ssh_message_free(message);
  } while (session->alive && chan == NULL);

  if(!session->alive) {
    return -1;
  }
  
  return 0;
}

int sshchan::connect(const sockaddr_in &dst, ssh_cleanup_frame **frames) {
    SSH_OPTIONS *opt = ssh_options_new();
    ssh_options_set_host(opt, inet_ntoa(dst.sin_addr));
    ssh_options_set_port(opt, ntohs(dst.sin_port));
    // TODO: Optionally set username
    // TODO: this should be different for servers, somehow?
    const char *username = rpc_primary_user();
    ssh_options_set_username(opt,username);
    const char *identloc = rpc_private_key_path(username);
    int oldstate;
    assert(pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &oldstate) == 0);
    ssh_options_set_identity(opt, identloc);
    int oldold = oldstate;
    assert(pthread_setcancelstate(oldold, &oldstate) == 0);
    if (wfs_dbg_block(3)) {
      ssh_options_set_log_verbosity(opt, sshloglevel);
    }
    // to make sure it uses non-blocking connect, so it can be canceled:
    ssh_options_set_timeout(opt, 60, 0);

    session = ssh_new();
    ssh_set_options(session,opt);
    if(ssh_connect(session, frames)){
      wfs_log(WFS_DBG_OFF, "ssh_connect : %s\n",
	      ssh_get_error(session));
      ssh_disconnect(session);
      session = NULL;
      return -1;
    }
    ssh_options_set_timeout(opt, 0, 0);

    // Auth: automatically use any available SSH keys
    // Currently, we assume an empty passphrase
    wfs_log(WFS_DBG_2, "sshchan connect trying autopubkey\n");
    int res = ssh_userauth_autopubkey(session, NULL, frames);
    if (res != SSH_AUTH_SUCCESS) {
      if (res == SSH_AUTH_ERROR) {
        wfs_log(WFS_DBG_OFF,  "sshchan connect auth error: %s\n", ssh_get_error(session));
      } else {
        wfs_log(WFS_DBG_OFF, "sshchan connect auth denied; is '%s' valid?\n",identloc);
      }
      return -1;
    }

    // check the host key here; we wait until after auth to make sure
    // we have the banner, and because doing pubkey to J. R. Malicious isn't
    // a security risk.
    char *pubkey = ssh_get_pubkey_base64(session);
    char *banner = ssh_get_issue_banner(session);
    char *primuser = banner; // strlen("SSH-2.0-WheelFS-"
    _cancel->cleanup_push( &free, pubkey );
    _cancel->cleanup_push( &free, primuser );
    wfs_log(WFS_DBG_2, "sshchan connect: about to check the host key against user %s...\n",primuser);
    if (!rpc_public_key_identifies_user(primuser,
					pubkey)) {
      wfs_log(WFS_DBG_OFF, "ssh_connect : unauthorized key for %s\n",primuser);
      ssh_disconnect(session);
      session = NULL;
      free(pubkey);
      free(primuser);
      _cancel->cleanup_pop( &free, primuser );
      _cancel->cleanup_pop( &free, pubkey );
      return -1;
    }
    _cancel->cleanup_pop( &free, primuser );
    _cancel->cleanup_pop( &free, pubkey );
    free(pubkey);
    strcpy(authorized_user,primuser);
    free(primuser);

    wfs_log(WFS_DBG_2, "sshchan connect auth succeeded; opening session\n");

    chan = channel_new(session);
    if (channel_open_session(chan, frames)) {
      wfs_log(WFS_DBG_OFF,
	      "sshchan open error: %s\n", ssh_get_error(session));
      return -1;
    }

    fd = ssh_get_fd(session);

    wfs_log(WFS_DBG_2, "sshchan connect opening session succeeded, channo=%d\n",fd);
    return 0;
}

// for debugging -- do something bad to the connection.
void
sshchan::lose()
{
  // TODO: Do a better job of implementing lossage for SSH
  /*
    int x = (random() % 3);
    if(x == 0){
    fprintf(stderr, "sshchan::lose shutdown read %d\n", channo());
    shutdown(s, SHUT_RD);
    } else if(x == 1){
    fprintf(stderr, "sshchan::lose shutdown write %d\n", channo());
    shutdown(s, SHUT_WR);
    } else {
  */
  wfs_log(WFS_DBG_OFF, "sshchan::lose close everything %d\n", channo());
  close();
  chan = NULL;
  //}
}

void sshchan::close() {
  if (chan != NULL) {
    channel_free(chan);
    chan = NULL;
  }
  if (session != NULL) {
    ssh_disconnect(session);
    session = NULL;
  }
}

void buff_cleanup( void *arg ) {
  BUFFER *b = (BUFFER *) arg;
  wfs_log(WFS_DBG_3, "(%x) sshchan::xread buff cleanup\n",
	  (unsigned) pthread_self() );
  buffer_free(b);
}

// read exactly n bytes.
int
sshchan::xread(void *xp, unsigned int n, ssh_cleanup_frame **frames)
{

  // TODO: is there any way to avoid a memory copy by reusing the buffer?
  BUFFER *buff = buffer_new();
  int cc;
  _cancel->cleanup_push( &buff_cleanup, (void *) buff );
  wfs_log(WFS_DBG_2, "(%x) xread temp: in xread %d\n",
		     (unsigned) pthread_self(), channo());
  cc = channel_read(chan, buff, n, false, frames);
  wfs_log(WFS_DBG_2, "(%x) xread temp: done with xread %d\n",
		     (unsigned) pthread_self(), channo());
  _cancel->cleanup_pop( &buff_cleanup, (void *) buff );
  if(cc < 0){
    wfs_log(WFS_DBG_OFF, "sshchan::xread read fd %d dead? %d: %s\n", channo(),
	    isdead?1:0, ssh_get_error(session));
    buffer_free(buff);
    die();
    return -1;
  }
  if(cc == 0){
    wfs_log(WFS_DBG_2, "sshchan: Bad xread of 0 bytes, channo %d\n", channo());
    buffer_free(buff);
    die();
    return -1;
  }
  // these used to be asserts, but if the other side fails mid-packet
  // it looks like libssh will return partial buffers
  if( (unsigned)cc == n && cc == buffer_get_len(buff) ) {
    memcpy(xp, buffer_get(buff), n);
  } else {
    wfs_log(WFS_DBG_OFF, "Badly-sized packet on channo %d, returning error\n",
	    channo());
    buffer_free(buff);
    return -1;
  }

  buffer_free(buff);

  //if (debug) fprintf(stderr,"xread temp: from xread %d\n",channo());

  return 0;
}

bool
sshchan::dead() {
  if (isdead) return true;
  //  if (session == NULL || !session->alive || chan == NULL || channel_is_closed(chan) || channel_is_eof(chan)) {
  if (channel_is_closed(chan) || channel_is_eof(chan)) {
    die();
    return true;
  } else {
    return false;
  }
}

const char *sshchan::getpeername() {
  if (nam[0] == '\0') {
    snprintf(nam,20,"%s:%d",
	     session->options->host, session->options->port);
  }
  return nam;
}

int sshchan::channo() {
  return fd;
}

ssize_t sshchan::xwritev(const struct iovec *iovptr, int iovcnt) {
  ssize_t written = 0;
  //if (debug) fprintf(stderr,"sshchan temp: in xwritev for %d\n",channo());
  for (int lcv = 0; lcv < iovcnt; lcv++) {
    int wr = channel_write(chan, iovptr[lcv].iov_base, iovptr[lcv].iov_len);
    if (wr <= 0) {
      //if (debug) fprintf(stderr,"sshchan temp: err xwritev for %d\n",channo());
      return wr;
    } else {
      assert ((unsigned)wr == iovptr[lcv].iov_len);
      written += wr;
    }
  }
  //if (debug) fprintf(stderr,"sshchan temp: ex xwritev for %d\n",channo());
  return written;
}

bool sshchan::verify_user(const char *username) {
  if (authorized_user[0] == '\0') {
    wfs_log(WFS_DBG_2, "sshchan::verify_user: an unauthorized channel somehow tried to act as '%s'!\n",username);
    return false;
  } else if (strcmp(authorized_user,"host") == 0
	     || strcmp(username, authorized_user) == 0) {
    return true;
  } else {
    wfs_log(WFS_DBG_2, "sshchan::verify_user: a channel authorized as '%s' tried to act as '%s'!\n",authorized_user,username);
    return false;
  }
}

sshchan *sshbind::accept(canceller *cancel,
			 pthread_mutex_t *sshdelete_m,
			 bool freebuf) {
  SSH_SESSION *session;

  session = ssh_bind_accept(bind);
  if(!session){
    wfs_log(WFS_DBG_OFF,
	    "ssh_bind_accept error: %s\n",
	    ssh_get_error(bind));
    return NULL;
  }
  return new sshchan(sshdelete_m,freebuf,cancel,
		     session,ssh_get_fd(session));
}

static void ssh_cleanup() {
  ssh_finalize();
}

sshhub::sshhub() {
  atexit(ssh_cleanup);
}

sshchan *sshhub::channel(canceller *cancel,
			 pthread_mutex_t *sshdelete_m,
			 bool freebuf)  {
  return new sshchan(sshdelete_m,freebuf,cancel);
}

sshbind *sshhub::bind(const sockaddr_in &sin) {
    SSH_BIND *bind = ssh_bind_new();

    SSH_OPTIONS *opt = ssh_options_new();
    ssh_options_set_bind(opt,
			 inet_ntoa(sin.sin_addr),
			 ntohs(sin.sin_port));
    //ssh_options_set_wanted_algos
    //ssh_options_set_dsa_server_key(opt,"/etc/ssh_host_dsa_key");
    ssh_options_set_rsa_server_key(opt,rpc_private_key_path(rpc_primary_user()));

    ssh_bind_set_options(bind, opt);

    int res = ssh_bind_listen(bind);
    if(res < 0){
      wfs_log(WFS_DBG_OFF, "ssh_bind_listen error: %s\n",ssh_get_error(bind));
      return NULL;
    }

    // Get port
    sockaddr_in actual;
    socklen_t actual_len = sizeof(actual);
    ::getsockname(ssh_bind_get_fd(bind), (struct sockaddr *) &actual,&actual_len);

    return new sshbind(bind, ntohs(actual.sin_port));
}

int sshhub::initial_timeout() {
  return 256000;
}


#endif

