/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// GC combined w. exceptions might make threads+locks+delete much easier.

// To Do:
// find out why freebsd runs out of threads, even in wanrpc server
//   maybe -lc_r instead of -lpthread?
// why does it burn up so much CPU? in wanrpc.
// tcp back-pressure to avoid huge inq

/**
 *
 * Summary of the memory conventions of using the RPC library:
 *
 * A marshalled message can be creating using the "marshall" class.
 * Upon instantiation, the object creates a memory buffer using
 * malloc.  Unless take_buf is called on the marshall object, the
 * memory is free'd when the marshall object is deleted or goes out of
 * scope.
 *
 * The marshall object can be converted into a sendable pdu by creating a
 * charbuf and passing it the marshall's memory using take_buf.  Note that
 * the size of the buffer needs to be saved before the call to take_buf.
 *
 * Because the user of a cchan or schan (e.g., rpcc or rpcs) may have
 * to send the same message again, the channels do not delete the
 * pdu's buffer after a successful send -- this is up to the user to
 * delete once the reply for the message has been received.  There are
 * two exceptions: 1) if a client's message has timed out; in this
 * case, if the charbuf's refcount is properly managed, cchan will
 * know it has the only reference left to the charbuf, and free the
 * buffer itself.  2) if a server is sending a reply that will not be
 * cached (like an atmostonce error); in this case, the charbuf's free
 * flag is set, and the tcpchan will delete the buffer after the
 * message is sent.
 *
 * Other conventions:
 *
 * Any thread blocking on a fifo deq operation needs to be canceled
 * before that fifo object is deleted.
 *
 **/

#include <sys/types.h>
#include <arpa/inet.h>
#include <signal.h>
#include "method_thread.h"
#include "chan.h"
#include "tcpchan.h"
#include "marshall.h"
#include <fcntl.h>
#include <sys/select.h>
#include <sys/uio.h>
#include "logging/wfs_log.h"

tcpchan::tcpchan(int xs, pthread_mutex_t *tcpdelete_m,
		 bool freebuf, canceller *cancel)
  : abstract_channel(tcpdelete_m, freebuf, cancel), s(xs)
{
}

tcpchan::~tcpchan()
{
  wfs_log(WFS_DBG_2, "~tcpchan socket %d\n", s );
  assert(pthread_cancel(th) == 0);
  assert(pthread_join(th, NULL) == 0);
  ::close(s);
  wfs_log(WFS_DBG_2, "~tcpchan done socket %d\n", s );
}

#ifdef WFS_WITH_LIBSSH
int tcpchan::connect(const sockaddr_in &dst, ssh_cleanup_frame **frames) {
#else
int tcpchan::connect(const sockaddr_in &dst) {
#endif
  int yes = 1;
  int ret;

  // we need to do the connect in non-blocking mode because connect
  // is not a cancellation point, but select is.
  long flags = fcntl( s, F_GETFL, NULL );
  flags |= O_NONBLOCK;
  fcntl( s, F_SETFL, flags );
  ret = ::connect(s, (sockaddr*)&dst, sizeof(dst));
  if( ret < 0 ) {
    if( errno == EINPROGRESS ) {
      fd_set fds;
      FD_ZERO(&fds);
      FD_SET( s, &fds );
      ret = select( s+1, NULL, &fds, NULL,
		    NULL /* no timeout, parent thread will cancel */ );
      if( ret < 0 && errno != EINTR ) {
	::close(s);
	perror("tcpchan connect/select");
	return ret;
      } else {
	int valopt;
	socklen_t sz = sizeof(int);
	getsockopt( s, SOL_SOCKET, SO_ERROR, (void *) &valopt, &sz );
	if( valopt != 0 ) {
	  ::close(s);
	  perror("tcpchan connect/select SO_ERROR");
	  wfs_log(WFS_DBG_OFF, "(error code was %d=%s)\n",valopt,strerror(valopt));
	  return valopt;
	}
      }
    }
  }

  // back to blocking
  flags = fcntl( s, F_GETFL, NULL );
  flags &= (~O_NONBLOCK);
  fcntl( s, F_SETFL, flags );

  setsockopt(s, IPPROTO_TCP, TCP_NODELAY, &yes, sizeof(yes));
  return 0;

}

// for debugging -- do something bad to the connection.
void
tcpchan::lose()
{
  int x = (random() % 3);
  if(x == 0){
    wfs_log(WFS_DBG_OFF, "tcpchan::lose shutdown read %d\n", s);
    shutdown(s, SHUT_RD);
  } else if(x == 1){
    wfs_log(WFS_DBG_OFF, "tcpchan::lose shutdown write %d\n", s);
    shutdown(s, SHUT_WR);
  } else {
    wfs_log(WFS_DBG_OFF, "tcpchan::lose close everything %d\n", s);
    ::close(s);
    s = -2;
  }
}

void tcpchan::close() {
  ::close(s);
}

// read exactly n bytes.
int
#ifdef WFS_WITH_LIBSSH
tcpchan::xread(void *xp, unsigned int n, ssh_cleanup_frame **frames)
#else
tcpchan::xread(void *xp, unsigned int n)
#endif
{
  char *p = (char *) xp;
  unsigned int i = 0;
  while(i < n){
    int cc = read(s, p + i, n - i);
    if(cc < 0){
      perror("tcpchan read");
      die();
      wfs_log(WFS_DBG_OFF, "tcpchan::xread read fd %d dead? %d\n", s, isdead?1:0);
      return -1;
    }
    if(cc == 0){
      wfs_log(WFS_DBG_2, "Bad xread of 0 bytes, channo %d\n", s);
      die();
      return -1;
    }
    i += cc;
  }
  return 0;
}

bool
tcpchan::dead() {
  if (isdead) return true;
  // do a fake read on the socket to make sure the connection is still alive
  char x;
  int stat = ::recv(s, &x, 1, MSG_PEEK|MSG_DONTWAIT);
  if (stat == 0 || (stat < 0 && errno != EAGAIN)) {
    wfs_log(WFS_DBG_2, "Connection closed (%d %d %d); dying\n", s, stat, errno);
    die();
    return true;
  } else {
    return false;
  }
}

const char *tcpchan::getpeername() {
   if (nam[0] == '\0') {
     sockaddr_in name;
     socklen_t namelen = sizeof(name);
     ::getpeername(s, (struct sockaddr *) &name, &namelen);
     snprintf(nam,20,"%s:%d",
	      inet_ntoa(name.sin_addr),
	       htons(name.sin_port));
   }
   return nam;
}

int tcpchan::channo() {
  return s;
}

ssize_t tcpchan::xwritev(const struct iovec *iovptr, int iovcnt) {
  return writev(s, iovptr, iovcnt);
}

bool tcpchan::verify_user(const char *username) {
  return true;
}

tcpbind::tcpbind(int _soc) : soc(_soc) {
  sockaddr_in name;
  socklen_t namelen = sizeof(name);
  ::getsockname(soc, (struct sockaddr *) &name, &namelen);
  port = ntohs(name.sin_port);
}

tcpchan *tcpbind::accept(canceller *cancel,
			 pthread_mutex_t *tcpdelete_m,
			 bool freebuf) {
  sockaddr_in sin;
  socklen_t slen = sizeof(sin);
  int s1 = -1;
  do {
    s1 = ::accept(soc, (sockaddr*)&sin, &slen);
    if( s1 < 0 ) {
      perror("tcpbind couldn't accept");
    }
  } while( s1 < 0 );

  int yes = 1;
  setsockopt(s1, IPPROTO_TCP, TCP_NODELAY, &yes, sizeof(yes));

  return new tcpchan(s1,tcpdelete_m,freebuf,cancel);
}


tcpchan *tcphub::channel(canceller *cancel,
			 pthread_mutex_t *tcpdelete_m,
			 bool freebuf) {
  int soc = ::socket(AF_INET, SOCK_STREAM, 0);
  if (soc < 0) {
    perror("tcpchan socket");
    return NULL;
  }
  return new tcpchan(soc,tcpdelete_m,freebuf,cancel);
}

tcpbind *tcphub::bind(const sockaddr_in &sin) {
  int soc = ::socket(AF_INET, SOCK_STREAM, 0);
  if (soc < 0)
    return NULL;
  int yes = 1;
  setsockopt(soc, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
  setsockopt(soc, IPPROTO_TCP, TCP_NODELAY, &yes, sizeof(yes));
  int ret = ::bind(soc, (sockaddr *)&sin, sizeof(sin));
  if (ret < 0) {
    perror("schan tcp bind");
    return NULL;
  }

  ret = ::listen(soc, 1000);
  if (ret < 0) {
    perror("schan tcp listen");
    return NULL;
  }
  return new tcpbind(soc);
}

int tcphub::initial_timeout() {
  return 128000;
}
