/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * stolen from DHASH chord/server.C -- jinyang
 *
 * Copyright (C) 2000 Frans Kaashoek (kaashoek@lcs.mit.edu)
 * Copyright (C) 2001 Frans Kaashoek (kaashoek@lcs.mit.edu) and
 *                    Frank Dabek (fdabek@lcs.mit.edu).
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */


#include <assert.h>
#include <math.h>

#include <wfs_misc.h>
#include "logging/wfs_log.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>

#include "vivaldi.h"


vivaldi::vivaldi()
{
  assert(pthread_mutex_init(&_m, 0) == 0);
}

vivaldi::~vivaldi()
{
  assert(pthread_mutex_destroy(&_m) == 0);

}

double
vivaldi::gettime()
{
#ifdef __APPLE__
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return (((double)tv.tv_sec)*1000000.0+(double)tv.tv_usec);
#else
  timespec ts;
  clock_gettime(CLOCK_REALTIME, &ts);
  return (((double)ts.tv_sec*1000000.0)+(double)(ts.tv_nsec/1000.0));
#endif
}

std::string
vivaldi::print_allcoords()
{
  int bufsize = 10000;
  int buflimit = bufsize-100;
  char buf[bufsize];
  bzero(buf, bufsize);
  char *buff = &buf[0];
  std::string big = "";
  for (__gnu_cxx::hash_map< unsigned, coord_info>::iterator i = _info.begin(); i!=_info.end(); ++i) {
    struct in_addr in;
    in.s_addr = (*i).first ^ ((*i).second.port << 16);
    buff += sprintf(buff, "%s:%d,%.2f,%.2f (%s) ", inet_ntoa(in), ntohs((*i).second.port),
		    _coords.distance_f((*i).second.c),
		    _lat.find(in.s_addr)==_lat.end()?-1.0:_lat[in.s_addr],
		    (*i).second.c.print_coord().c_str() );
    if( (buff-&buf[0]) > buflimit ) {
      big += std::string(buf);
      bzero(buf, bufsize);
      buff = &buf[0];
    }
  }
  return (big + std::string(buf));
}

bool
vivaldi::get_coords(unsigned long long k , Coord &c)
{
  if (_info.find(k) == _info.end()) {
    return false;
  }else{
    c = _info[k].c;
    return true;
  }
}

void
vivaldi::update_other_coords (uint32_t ip, uint16_t port, Coord uc, double age)
{
  double now = gettime();
  unsigned k = ip ^ (port << 16);
  if (_info.find(k) == _info.end() ||
      (now + _info[k].when + _info[k].age) >= age) {
    _info[k].c = uc;
    _info[k].when = now;
    _info[k].age = age;
    _info[k].port = port;
  }
}

void
vivaldi::update_other_lat (uint32_t ip, double lat)
{
  if (_lat.find(ip) == _lat.end()) {
    _lat[ip] = lat;
  }else{
    _lat[ip] = 0.95 * _lat[ip] + 0.05 * lat; //smooth it out with EWMA
  }
}

double
vivaldi::get_distance_to_ip (uint32_t ip, uint16_t port)
{
  if (_lat.find(ip)!=_lat.end()) {
    return _lat[ip];
  }else {
    unsigned k = ip ^ (port << 16);
    if (_info.find(k)!=_info.end())
      return _coords.distance_f(_info[k].c);
    else
      return 1000000; /* one second default */
  }
}

void
vivaldi::update_coords (Coord uc, double ud)
{
  assert(pthread_mutex_lock(&_m) == 0);

  //  warn << myID << " --- starting update -----\n";
  Coord coords = _coords;
  Coord v = uc;
  double rmt_err = uc.err ();


  double actual = ud;
  double expect = coords.distance_f (uc);


  if (actual >= 0.0 && actual < 1000000) { //ignore timeouts

    //track our prediction error
    update_error (actual, expect, rmt_err);

    // force magnitude: > 0 --> stretched
    double grad = expect - actual;
    v.vector_sub (coords);

    double len = v.plane_norm ();
    while (len < 0.0001) {
      for (int i = 0; i < NCOORD; i++)
	v.coords[i] = (double)(random () % 400 - 200) / 1.0;
      //if (USING_HT) v.ht += fabs((double)(random () % 10 - 5) / 10.0);
      len = v.plane_norm ();
    }
    len = v.norm ();

    double unit = 1.0/sqrtf(len);

    // scalar_mult(v, unit) is unit force vector
    // times grad gives the scaled force vector
    v.scalar_mult (unit*grad);


    //timestep is the scaled ratio of our prediction error
    // and the remote node's prediction error
    double pred_err = _coords.err ();
    double timestep;
    if (pred_err > 0 && rmt_err > 0)
      timestep = 0.1 * (pred_err)/(pred_err + rmt_err);
    else if (pred_err > 0)
      timestep = pred_err * 0.1; //crazy modifcation by jy
    else
      timestep = 1.0;

    v.scalar_mult(timestep);
    //flip sign on height
    v.ht = -v.ht;

    coords.vector_add (v);

#if 0
    wfs_log(WFS_DBG_OFF,"coord hop: %f - %f = %f ; len=%f ts=%f (%f %f)\n",
			 expect, actual, grad, len, timestep, pred_err, rmt_err);
    wfs_log(WFS_DBG_OFF, "v : %s ----------\n", v.print_coord().c_str());
    wfs_log(WFS_DBG_OFF, "Me: old %s new %s\n", _coords.print_coord().c_str(), coords.print_coord().c_str());
#endif
    if (coords.ht <= 100) coords.ht = 100;
	 int save_pred_err = _coords.raw_err();
    _coords = coords;
	 _coords.set_raw_err(save_pred_err);
  } else {
    wfs_log(WFS_DBG_3, "COORD:  ignored actual of %f\n", actual);
  }
  assert(pthread_mutex_unlock(&_m) == 0);
}

void
vivaldi::update_error (double actual, double expect, double rmt_err)
{
  if (actual < 0) return;
  double rel_error = fabs (expect - actual)/actual;


  double pred_err = _coords.err ();
  double npe = -1.0;

  if (pred_err < 0)
    npe = rel_error;
  else if (rmt_err < 0)
    npe = pred_err;
  else {
    // ce is our pred error, he is the remote pred error
    // squaring them punishes high error nodes relatively more.
    double ce = pred_err*pred_err;
    double he = rmt_err*rmt_err;
    //this is our new prediction error found by combining our prediction
    // error with the remote node's error
    double new_pred_err = rel_error*(ce/(he + ce)) + pred_err*(he/(he + ce));

    //don't just take the new prediction, EWMA it with old predictions
    npe = (19*pred_err + new_pred_err)/20.0;

    //cap it at 1.0
    if (npe > 1.0) npe = 1.0;

  }
  _coords.update_err(npe);

}


#define SET_COORDS(NC, val)               \
   coords.clear ();                         \
   for (unsigned int i = 0; i < NC; i++) { \
      coords.push_back (val);               \
   }                                       \

Coord::Coord ()
{
  coords.clear();
  for (int i = 0; i < NCOORD; i++)
	 coords.push_back((double)random()*10.0/RAND_MAX);
  ht = 1.0;
  pred_err = -1;
};

void
Coord::set (const Coord &c)
{
  SET_COORDS (c.size (), c.coords[i]);
  if (USING_HT) ht = c.ht;
  pred_err = c.pred_err;
}

void
Coord::set_coord(int i, double x)
{
  assert(i < NCOORD);
  coords[i] = x;
}

double
Coord::get_coord(int i)
{
  assert(i < NCOORD);
  return coords[i];
}

std::string
Coord::print_coord ()
{
  char buf[1000];
  char *b = &buf[0];
  for (unsigned int i = 0; i < coords.size (); i++)
    b += sprintf(b, "%d ", (int)(coords[i]));
  if (USING_HT) b+= sprintf(b,"ht=%d ",(int)ht);
  sprintf(b, "with scaled error: %.3f ", err());
  return std::string(buf);
}



double
Coord::distance_f (const Coord &c)
{
  double f = 0.0;
  for (unsigned int i = 0; i < coords.size (); i++)
    f += (c.coords[i] - coords[i])*(c.coords[i] - coords[i]);
  f = sqrt (f);
  if (USING_HT)
    f += ht + c.ht;
  return f;
}

double
Coord::distance_f (const Coord &a, const Coord &b)
{

  double f = 0.0;
  for (unsigned int i = 0; i < a.size (); i++)
    f += (a.coords[i] - b.coords[i])*(a.coords[i] - b.coords[i]);
  f = sqrt (f);
  if (USING_HT)
    f += a.ht + b.ht;
  return f;
}

void
Coord::vector_add (const Coord &b)
{
  for (unsigned int i = 0; i < coords.size (); i++)
    coords[i] = (coords[i] +  b.coords[i]);
  ht += b.ht;
}

void
Coord::vector_sub (const Coord &b)
{
  for (unsigned int i = 0; i < coords.size (); i++)
    coords[i] = (coords[i] - b.coords[i]);
  ht += b.ht;
}

double
Coord::norm ()
{
  double ret = 0.0;
  for (unsigned int i = 0; i < coords.size (); i++)
    ret += coords[i]*coords[i];
  if (USING_HT)
    ret += ht;
  return ret;
}

double
Coord::plane_norm ()
{
  double ret = 0.0;
  for (unsigned int i = 0; i < coords.size (); i++)
    ret += coords[i]*coords[i];
  return ret;
}


void
Coord::scalar_mult (double s)
{
  for (unsigned int i = 0; i < coords.size (); i++)
    coords[i] = (coords[i]*s);
  ht *= s;
}

