/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// WAN RPC test program. time how long it takes to
// send lots of RPC data, to make sure rpc-over-tcp
// is actually doing some good.
//
// host1% wanrpc 9999
// host2% wanrpc host1 9999

#include <sys/time.h>
#include <time.h>
#include "rpc.h"
#include <arpa/inet.h>
#include <assert.h>

class srv {
public:
  int handle_22(const std::string a, std::string & r);
};

int
srv::handle_22(const std::string a, std::string &r)
{
  r = a;
  return 0;
}

void
server(int port)
{
  rpcs server(port);
  srv service;
  server.reg(22, &service, &srv::handle_22);
  while(1)
    sleep(100);
}

rpcc *c;
struct timeval tv0, tv1;

void
tstart()
{
  gettimeofday(&tv0, 0);
}

double
tend()
{
  gettimeofday(&tv1, 0);
  double t = (tv1.tv_sec - tv0.tv_sec) +
    (tv1.tv_usec - tv0.tv_usec) / 1000000.0;
  return t;
}

void
one(unsigned n)
{
  std::string req((size_t) n, 'x');
  std::string rep;
  int ret = c->call(22, req, rep);
  assert(ret == 0);
  assert(rep == req);
  assert(rep.size() == n);
}

void *
client1(void *xx)
{
  int n = (int)(long) xx;
  one(n);
  return NULL;
}

bool client2_done;
void *
client2(void *xx)
{
  time_t t0;
  time(&t0);
  while(time(0) < t0 + 5){
    one(200000);
  }
  client2_done = 1;
  return NULL;
}

void
client(char *host, int port)
{

  sockaddr_in dst;
  bzero(&dst, sizeof(dst));
  dst.sin_family = AF_INET;
  dst.sin_addr.s_addr = inet_addr(host);
  dst.sin_port = port;

  c = new rpcc(dst);

  // cause the tcp connection to be open
  one(1000);

  // small RPC latency
  tstart();
  for(int i = 0; i < 10; i++)
    one(1);
  printf("small RPC latency %.4f seconds\n", tend() / 10);

  // one-K RPC latency
  tstart();
  for(int i = 0; i < 10; i++)
    one(1024);
  printf("one-K RPC latency %.4f seconds\n", tend() / 10);

  // 500 small concurrent RPCs
  {
    int nt = 500, ret;
    pthread_t th[nt];
    tstart();
    for(int i = 0; i < nt; i++){
      ret = pthread_create(&th[i], NULL, client1, (void *) 1);
      if(ret != 0){
        fprintf(stderr, "pthread_create ret %d %s\n", ret, strerror(ret));
        exit(1);
      }
    }
    for(int i = 0; i < nt; i++){
      assert(pthread_join(th[i], NULL) == 0);
    }
    printf("%d concurrent small in %.3f\n", nt, tend());
  }

  // 500 one-K concurrent RPCs
  {
    int nt = 500, ret;
    pthread_t th[nt];
    tstart();
    for(int i = 0; i < nt; i++){
      ret = pthread_create(&th[i], NULL, client1, (void *) 1024);
      if(ret != 0){
        fprintf(stderr, "pthread_create ret %d %s\n", ret, strerror(ret));
        exit(1);
      }
    }
    for(int i = 0; i < nt; i++){
      assert(pthread_join(th[i], NULL) == 0);
    }
    printf("%d concurrent one-K in %.3f %.2fM/s\n",
           nt,
           tend(),
           (nt * 1024 / 1000000.0) / tend());
  }

  // throughput of sequential large RPCs
  tstart();
  double total = 0;
  while(tend() < 10){
    one(100000);
    total += 100000;
  }
  printf("%.1fM at %.3f MB/s\n",
         total / 1000000,
         (total / 1000000) / tend());

  // small RPC while one huge RPC is going,
  // hoping small RPC will go via UDP.
  {
    // start big rpc
    client2_done = false;
    pthread_t bigth;
    int ret = pthread_create(&bigth, NULL, client2, 0);
    if(ret != 0){
      fprintf(stderr, "pthread_create ret %d %s\n", ret, strerror(ret));
      exit(1);
    }

    usleep(1000);

    tstart();
    int n = 0;
    while(client2_done == false){
      one(1);
      n++;
    }
    printf("small latency during big xfer %.3f\n", tend() / n);

    assert(pthread_join(bigth, NULL) == 0);
  }
}

int
main(int argc, char *argv[])
{
  if(argc == 2){
    server(htons(atoi(argv[1])));
  } else if(argc == 3){
    client(argv[1], htons(atoi(argv[2])));
  } else {
    fprintf(stderr, "Usage: wanrpc port\n");
    fprintf(stderr, "       wanrpc host port\n");
    exit(1);
  }
}
