/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <assert.h>
#include "wfs_client_manager.h"
#include "wfs_misc.h"

const unsigned wfs_client_manager::CLIENT_LIMIT = 50;

wfs_client_manager::wfs_client_manager( vivaldi *viv )
  : _viv(viv), _head(NULL), _tail(NULL) {
  pthread_mutex_init( &_clients_lock, NULL );
}

wfs_client_manager::~wfs_client_manager() {
  pthread_mutex_destroy( &_clients_lock );
}

bool wfs_client_manager::delete_client( client_node *cnode ) {

  // only delete those that aren't out
  if( cnode->out != 0 ) {
    return false;
  }

  // if it's LRU, hopefully there are no outstanding calls, but
  // just in case, cancel them
  cnode->client->cancel();
  delete cnode->client;
  if( cnode->prev != NULL ) {
    cnode->prev->next = cnode->next;
  }
  if( cnode->next != NULL ) {
    cnode->next->prev = cnode->prev;
  }

  if( cnode->id == _head->id ) {
    _head = cnode->next;
  }

  if( cnode->id == _tail->id ) {
    _tail = cnode->prev;
  }

  _clients[cnode->id] = NULL;
  _clients.erase(cnode->id);
  delete cnode;
  return true;

}

rpcc *wfs_client_manager::get_client( const wfs_node &server, rpcc::TO to ) {

  wfs_log(WFS_DBG_2, "Getting client for node %016"PRIx64"\n", server.id );

  rpcc *client = NULL;
  client_node *cnode = NULL;
  pthread_mutex_lock( &_clients_lock );

  if( _clients.find( server.id ) != _clients.end() &&
      _clients[server.id]->client->dead() ) {
    wfs_log(WFS_DBG_1, "Client to node %016"PRIx64" died; retrying\n", server.id );
    cnode = _clients[server.id];
    delete_client(cnode);
  }

  // if there's a binding in progress, wait for it
  while( _bindings.find( server.id ) != _bindings.end() &&
	 _bindings[server.id]->waiters >= 0 &&
	 _clients.find( server.id ) == _clients.end() ) {

    _bindings[server.id]->waiters++;
    pthread_cond_wait( _bindings[server.id]->cond, &_clients_lock );
    if( _bindings[server.id]->waiters > 0 ) {
      _bindings[server.id]->waiters--;
    } else {
      _bindings[server.id]->waiters++;
    }

    if( _bindings[server.id]->waiters == 0 ) {
      binding *bind = _bindings[server.id];
      _bindings[server.id] = NULL;
      _bindings.erase(server.id);
      assert( pthread_cond_destroy( bind->cond ) == 0 );
      delete bind->cond;
      delete bind;
    }

  }

  if( _clients.find( server.id ) == _clients.end() ) {

    client = new rpcc(server.addr);
    if( _viv != NULL ) {
      client->set_vivaldi(_viv);
    }

    binding *bind;
    if( _bindings.find(server.id) != _bindings.end() ) {
      bind = _bindings[server.id];
      assert( bind->waiters < 0 );
      bind->waiters = -(bind->waiters); // flip it back
    } else {
      bind = new binding();
      _bindings[server.id] = bind;
      bind->waiters = 0;
      bind->cond = new pthread_cond_t();
      assert( pthread_cond_init( bind->cond, NULL ) == 0 );
    }
    bind->waiters++;

    // bind first, ask questions later
    pthread_mutex_unlock( &_clients_lock );
    int status = client->bind(to);
    pthread_mutex_lock( &_clients_lock );

    bind->waiters--;
    if( bind->waiters == 0 ) {
      _bindings[server.id] = NULL;
      _bindings.erase(server.id);
      assert( pthread_cond_destroy( bind->cond ) == 0 );
      delete bind->cond;
      delete bind;
    } else {
      if( status != 0 ) {
	// flip the waiters sign to negative so the waiters know we've given up
	bind->waiters = -(bind->waiters);
      }
      pthread_cond_broadcast( bind->cond );
    }

    if( status != 0 ) {
      cnode = NULL;
      wfs_log(WFS_DBG_1, "Bind to node %016"PRIx64" failed (%d), exiting\n",
          server.id, status );
      delete client;
      client = NULL;
      goto exit;
    }

    // if we have too many TCP connections open already, must destroy
    // the LRU connection
    // TODO: prefer to kick out dead connections?
    if( _clients.size() >= CLIENT_LIMIT ) {
      if( _tail != NULL ) {
        wfs_log(WFS_DBG_2, "Deleting old client for node %016"PRIx64"\n", _tail->id );
        client_node *to_delete = _tail;
        while( to_delete != NULL && !delete_client(to_delete) ) {
          to_delete = to_delete->prev;
        }
      }
    }

    cnode = new client_node();
    cnode->id = server.id;
    cnode->client = client;
    cnode->prev = NULL;
    cnode->next = _head;
    cnode->out = 1;
    if( _head != NULL ) {
      _head->prev = cnode;
    }
    if( _tail == NULL ) {
      _tail = cnode;
    }
    _head = cnode;

    _clients[server.id] = cnode;

  } else {

    cnode = _clients[server.id];
    client = cnode->client;
    cnode->out++;

    // set as MRU
    if( cnode->id != _head->id ) {

      if( cnode->id == _tail->id ) {
	_tail = cnode->prev;
      }

      if( cnode->prev != NULL ) {
	cnode->prev->next = cnode->next;
      }
      if( cnode->next != NULL ) {
	cnode->next->prev = cnode->prev;
      }
      _head->prev = cnode;
      cnode->next = _head;
      cnode->prev = NULL;
      _head = cnode;

    }

  }

 exit:

  pthread_mutex_unlock( &_clients_lock );
  return client;

}

void wfs_client_manager::done_with_client( const wfs_node &server ) {

  wfs_log(WFS_DBG_2, "Done with client for node %016"PRIx64"\n", server.id );

  pthread_mutex_lock( &_clients_lock );
  assert( _clients.find(server.id) != _clients.end() );
  _clients[server.id]->out--;
  pthread_mutex_unlock( &_clients_lock );

}
