/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <sys/time.h>
#include <sys/errno.h>

#include "wfs_perfile_locker.h"
#include "wfs_misc.h"

wfs_perfile_locker::wfs_perfile_locker() {
  assert( pthread_mutex_init( &_locks_mutex, NULL ) == 0 );
  assert( pthread_mutex_init( &_stale_mutex, NULL ) == 0 );
  assert( pthread_cond_init( &_stalewait_cv, NULL ) == 0 );
}

wfs_perfile_locker::~wfs_perfile_locker() {
  // TODO: delete any existing lock structs in _locks?
  pthread_mutex_destroy( &_locks_mutex );
  pthread_mutex_destroy( &_stale_mutex );
  pthread_cond_destroy( &_stalewait_cv );
}

void wfs_perfile_locker::lock_file( wfs_file_handle fh ) {

  lock_struct *lock;

  wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);

  pthread_mutex_lock( &_locks_mutex );
  lock = _locks[fid];
  if( lock == NULL ) {
    if( _new_locks.size() > 0) {
      lock = _new_locks.back();
      _new_locks.pop_back();
    } else {
      lock = new lock_struct();
    }
    _locks[fid] = lock;
  }
  lock->count++;
  pthread_mutex_unlock( &_locks_mutex );

  // should be ok to do the lock outside of the critical section,
  // because count has been incremented.

  wfs_log(WFS_DBG_2, "About to lock %016"PRIx64"\n", fh.fid );
  pthread_mutex_lock( lock->lock );
  wfs_log(WFS_DBG_2, "Locked %016"PRIx64"\n", fh.fid );
}

void wfs_perfile_locker::unlock_file( wfs_file_handle fh ) {

  lock_struct *lock;

  wfs_file_id fid = fh.fid; //file_handle_to_file_id(fh);

  pthread_mutex_lock( &_locks_mutex );
  lock = _locks[fid];
  assert(lock);
  lock->count--;
  if( lock->count == 0 ) {
    _new_locks.push_back(lock);
    _locks[fid] = NULL;
  }

  // I think we need to unlock the lock before unlocking the global lock,
  // in case the lock gets reused somewhere.
  pthread_mutex_unlock( lock->lock );

  pthread_mutex_unlock( &_locks_mutex );
  wfs_log(WFS_DBG_2, "Unlocked %016"PRIx64"\n", fh.fid );
}

void wfs_perfile_locker::mark_stale( wfs_file_handle fh, bool stale ) {

  pthread_mutex_lock( &_stale_mutex );

  _stale_map[fh.fid] = stale;

  // signal anyone waiting for this to go false
  if( stale == false && _stalewait_map[fh.fid] == true ) {
    _stalewait_map[fh.fid] = false;
    // because I'm lazy, we just share one condition variable for all
    // files.  The waiters must check the status of their file
    // when they wake up.
    pthread_cond_broadcast( &_stalewait_cv );
  }

  pthread_mutex_unlock( &_stale_mutex );

}

bool wfs_perfile_locker::is_stale( wfs_file_handle fh, int to_wait_ms ) {

  bool answer = false;
  pthread_mutex_lock( &_stale_mutex );

  if( _stale_map.find(fh.fid) != _stale_map.end() ) {
    answer = _stale_map[fh.fid];
  }

  // if it is stale, maybe wait for a while (or forever)
  if( answer == true && to_wait_ms >= 0 ) {

    _stalewait_map[fh.fid] = true;

    struct timeval now;
    gettimeofday(&now, NULL);
    struct timespec deadline;
    // convert to millisec, add timeout, convert back
    long double msec = (now.tv_sec*1000.0 + now.tv_usec/1000.0) +
      to_wait_ms*1.0;
    deadline.tv_sec = (long) (msec/1000);
    deadline.tv_nsec =
      ((long) ((msec-((long double) deadline.tv_sec*1000.0))*1000))*1000;

    // loop until we timeout or it's no longer stale
    bool repeat = true;
    do {

      if( to_wait_ms == 0 ) {
	// wait forever
	pthread_cond_wait( &_stalewait_cv, &_stale_mutex );
      } else {

	int ret = pthread_cond_timedwait( &_stalewait_cv, &_stale_mutex,
					  &deadline );
	if( ret == ETIMEDOUT ) {
	  repeat = false;
	}

      }

      answer = _stale_map[fh.fid];

    } while( answer == true && repeat );

  }

  pthread_mutex_unlock( &_stale_mutex );

  return answer;

}
